@@ -54,6 +54,35 @@ iscontrol (c)
   }
 }
 
+/**
+   Check if character is any of the following white space characters:
+	U+0009;<control>;Cc;0;S;;;;;N;CHARACTER TABULATION;;;;
+	U+0020;SPACE;Zs;0;WS;;;;;N;;;;;
+	U+00A0;NO-BREAK SPACE;Zs;0;CS;<noBreak> 0020;;;;N;NON-BREAKING SPACE;;;;
+	U+2002;EN SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2003;EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2004;THREE-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2005;FOUR-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2006;SIX-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2007;FIGURE SPACE;Zs;0;WS;<noBreak> 0020;;;;N;;;;;
+	U+2008;PUNCTUATION SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+2009;THIN SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+200A;HAIR SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+200B;ZERO WIDTH SPACE;Cf;0;BN;;;;;N;;;;;
+	U+202F;NARROW NO-BREAK SPACE;Zs;0;CS;<noBreak> 0020;;;;N;;;;;
+	U+205F;MEDIUM MATHEMATICAL SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
+	U+3000;IDEOGRAPHIC SPACE;Zs;0;WS;<wide> 0020;;;;N;;;;;
+ */
+int
+iswhitespace (c)
+  unsigned long c;
+{
+  return c == ' ' || c == '\t' || c == 0xA0
+	|| (c >= 0x2002 && c <= 0x200B) || c == 0x3000
+	|| c == 0x202F || c == 0x205F;
+}
+
+
 /**
    Return display indication for a control character.
  */
@@ -89,8 +118,32 @@ isolated_alef (unichar)
 		/* ALEF */
 		return 0xFE8D;
 	} else {
-		/* ? -> ALEF SYMBOL */
-		return 0x2135;
+		return 0xFE8D;
+	}
+}
+
+/**
+   Return the ligature with LAM of an ALEF character.
+   Use the ISOLATED FORM.
+ */
+unsigned long
+ligature_lam_alef (unichar)
+  unsigned long unichar;
+{
+	if (unichar == 0x0622) {
+		/* ALEF WITH MADDA ABOVE */
+		return 0xFEF5;
+	} else if (unichar == 0x0623) {
+		/* ALEF WITH HAMZA ABOVE */
+		return 0xFEF7;
+	} else if (unichar == 0x0625) {
+		/* ALEF WITH HAMZA BELOW */
+		return 0xFEF9;
+	} else if (unichar == 0x0627) {
+		/* ALEF */
+		return 0xFEFB;
+	} else {
+		return 0xFEFB;
 	}
 }
 
@@ -116,6 +169,8 @@ max_char_value ()
 
 /**
    Convert CJK character in current text encoding to byte sequence.
+   Terminate with NUL byte.
+   Return byte count.
  */
 int
 cjkencode (cjkchar, buf)
@@ -183,7 +238,9 @@ cjkencode_char (term, cjkchar, buf)
 }
 
 /**
-   convert Unicode character to UTF-8
+   Convert Unicode character to UTF-8.
+   Terminate with NUL byte.
+   Return byte count.
  */
 int
 utfencode (unichar, buf)
@@ -234,6 +291,8 @@ utfencode (unichar, buf)
 
 /**
    Convert character to byte sequence.
+   Terminate with NUL byte.
+   Return buffer pointer.
  */
 char *
 encode_char (c)
@@ -399,8 +458,7 @@ struct charmap_table_entry {
 	char tag1;
 };
 
-static struct charmap_table_entry charmaps_table [] =
-{
+static struct charmap_table_entry charmaps_table [] = {
 #ifdef __TURBOC__
 	{cp437_table, & cp437_table_len, "CP437", "PC", 'p'},
 	{cp850_table, & cp850_table_len, "CP850", "PL", 'P'},
@@ -429,8 +487,8 @@ matchprefix (s, m)
   char * s;
   char * m;
 {
-  char cs, cm;
   do {
+	char cs, cm;
 	while (* m == '-' || * m == '_' || * m == ' ') {
 		m ++;
 	}
@@ -519,7 +577,7 @@ static unsigned int text_table_len = 0;
 
 #ifndef use_cjk_tables
 /* define dummy tables referred in code */
-#ifdef __TURBOC__
+#if defined(__TURBOC__) || defined(vms)
 #define EMPTY 1
 #else
 #define EMPTY 0
@@ -605,11 +663,8 @@ FLAG
 encoding_has_combining ()
 {
   return utf8_text
-	|| (cjk_text && 
-		(text_encoding_tag == 'G'
-		 || text_encoding_tag == 'J'
-		 || text_encoding_tag == 'S'))
-	|| (mapped_text && combined_text);
+	|| (mapped_text && combined_text)
+	|| (cjk_text && combined_text);
 }
 
 /**
@@ -716,8 +771,9 @@ setup_mapping (term, map_table, map_table_len, tag1, tag2)
 	if (multi_byte) {
 		cjk_text = True;
 		mapped_text = False;
-		/* combined_text is dynamically enquired with 
-		   encoding_has_combining () */
+		combined_text = text_encoding_tag == 'G'
+				|| text_encoding_tag == 'J'
+				|| text_encoding_tag == 'S';
 	} else {
 		mapped_text = True;
 		cjk_text = False;
@@ -737,8 +793,6 @@ set_char_encoding (term, charmap, tag)
   char * charmap;
   char tag;
 {
-  int i;
-
   if (charmap && ! term
 	&& (streq (":16", charmap) || matchpart ("UTF-16BE", charmap))) {
 	utf8_text = True;
@@ -764,6 +818,7 @@ set_char_encoding (term, charmap, tag)
 	text_encoding_tag = ' ';
 	text_encoding_flag = "??";
 	utf8_text = False;
+	utf16_file = False;
 	cjk_text = True;
 	mapped_text = False;
 	current_text_encoding = "[CJK]";
@@ -795,6 +850,7 @@ set_char_encoding (term, charmap, tag)
 		term_encoding_tag = 'L';
 	} else {
 		utf8_text = False;
+		utf16_file = False;
 		cjk_text = False;
 		mapped_text = False;
 		current_text_encoding = "ISO 8859-1";
@@ -802,6 +858,7 @@ set_char_encoding (term, charmap, tag)
 	}
 	return True;
   } else {
+    int i;
     for (i = 0; i < arrlen (charmaps_table); i ++) {
 	if (charmap ? (charmap [0] == ':' ?
 			  streq (& charmap [1], charmaps_table [i].tag2)
@@ -814,6 +871,7 @@ set_char_encoding (term, charmap, tag)
 			term_encoding = charmaps_table [i].charmap;
 		} else {
 			utf8_text = False;
+			utf16_file = False;
 			current_text_encoding = charmaps_table [i].charmap;
 		}
 		setup_mapping (term, 
@@ -828,12 +886,10 @@ set_char_encoding (term, charmap, tag)
   return False;
 }
 
-static
-struct {
+static struct {
 	char * alias;
 	char * codepage;
-} cpaliases [] =
-{
+} cpaliases [] = {
 	{"CP819", "ISO-8859-1"},
 	{"CP912", "ISO-8859-2"},
 	{"CP913", "ISO-8859-3"},
@@ -852,10 +908,24 @@ struct {
 	{"CP28595", "ISO-8859-5"},
 	{"CP28596", "ISO-8859-6"},
 	{"CP28597", "ISO-8859-7"},
-	{"CP28598", "ISO-8859-8"},
+	{"CP28598", "ISO-8859-8"},	/* indicates visual ordering ... */
 	{"CP28599", "ISO-8859-9"},
 	{"CP28603", "ISO-8859-13"},
 	{"CP28605", "ISO-8859-15"},
+	{"CP38598", "ISO-8859-8"},	/* indicates logical ordering ... */
+	{"CP20000", "CNS"},
+	{"CP20127", "ASCII"},
+	{"CP20866", "KOI8-R"},
+	{"CP20936", "GB2312"},
+	{"CP21866", "KOI8-U"},
+	{"CP51949", "EUC-KR"},
+	{"CP54936", "GB18030"},
+	{"CP65001", "UTF-8"},
+	{"CP932", "Shift-JIS"},
+	{"CP936", "GBK"},
+	{"CP949", "EUC-KR"},
+	{"CP950", "Big5"},
+	{"CP20932", "EUC-JP"},
 };
 
 /**
@@ -868,7 +938,7 @@ set_term_encoding (charmap, tag)
   char tag;
 {
   /* handle generic codepage notation (cygwin 1.7 / DOS support) */
-  if (strisprefix ("CP", charmap)) {
+  if (charmap && strisprefix ("CP", charmap)) {
 	if (set_char_encoding (True, charmap, tag)) {
 		return True;
 	} else {
@@ -1009,9 +1079,11 @@ gb_to_unicode (gb)
 	unsigned int byte2 = (gb >> 16) & 0xFF;
 	unsigned int byte3 = (gb >> 8) & 0xFF;
 	unsigned int byte4 = gb & 0xFF;
+
 	if (byte2 < '0' || byte2 > '9' || byte3 < 0x81 || byte4 < '0' || byte4 > '9') {
 		return CHAR_INVALID;
 	}
+
 	return (((((gb >> 24) & 0xFF) - 0x90) * 10
 		+ (byte2 - 0x30)) * 126L
 		+ (byte3 - 0x81)) * 10L
@@ -1025,6 +1097,11 @@ unicode_to_gb (uc)
   unsigned long uc;
 {
 	unsigned int a, b, c, d;
+
+	if (uc >= 0x200000) {
+		return CHAR_INVALID;
+	}
+
 	uc -= 0x10000;
 	d = 0x30 + uc % 10;
 	uc /= 10;
@@ -1033,6 +1110,7 @@ unicode_to_gb (uc)
 	b = 0x30 + uc % 10;
 	uc /= 10;
 	a = 0x90 + uc;
+
 	return (a << 24) | (b << 16) | (c << 8) | d;
 }
 
