@@ -63,8 +63,7 @@
 		A(-.		BREVE AND DOT BELOW
 		O9-.		HORN AND DOT BELOW
  */
-static struct prefixspec prefixmap [] =
-{
+static struct prefixspec prefixmap [] = {
 	/* function keys */
 	{F5, altshift_mask, "breve/vrachy", "˘", "x("},
 	{F5, ctrlshift_mask, "ogonek/prosgegrammeni", "˛", "x;"},
@@ -211,8 +210,7 @@ static
 struct {
 	char * mnemo;
 	int uni;
-} unimnemo3 [] =
-{
+} unimnemo3 [] = {
 #include "mnemos.h"
 	{NIL_PTR}
 };
@@ -337,7 +335,7 @@ subst_acc (c1p, c2p, acc, sub)
    Return CHAR_INVALID if character found is invalid (CJK conversion).
  */
 unsigned long
-compose (c1, c2)
+compose_chars (c1, c2)
   unsigned long c1;
   unsigned long c2;
 {
