@@ -187,6 +187,7 @@ DELkey ()
 	keyshift = 0;
 	CUT ();
   } else {
+#ifdef oldstyle_DELkey
 	/*
 	mined_keypad alt_mask applkeypad_mask	alt-KP	alt-DEL	alt-ALT
 		1	0	0		cut !	cut	cut
@@ -203,6 +204,21 @@ DELkey ()
 	} else {
 		DCC ();
 	}
+#else
+	if (has_active_selection ()) {
+		if (keyshift & alt_mask) {
+			DCC ();
+		} else {
+			CUT ();
+		}
+	} else {
+		if (keyshift & alt_mask) {
+			CUT ();
+		} else {
+			DCC ();
+		}
+	}
+#endif
   }
 }
 
@@ -273,6 +289,29 @@ smallINSkey ()
 }
 
 
+/**
+   Keypad keys with special functions
+ */
+void
+KP_plus ()
+{
+  if (keyshift & (ctrl_mask | shift_mask)) {
+	font_size (1);
+  } else {
+	SD ();
+  }
+}
+
+void
+KP_minus ()
+{
+  if (keyshift & (ctrl_mask | shift_mask)) {
+	font_size (-1);
+  } else {
+	SU ();
+  }
+}
+
 /**
    Display function key help lines
  */
@@ -354,7 +393,7 @@ F2 ()
   trace_dispatch ("F2");
 
   if (keyshift & ctrl_mask) {
-	/* enforce creation of marker position memory file @mined.mar */
+	/* enforce creation of file info file */
 	hop_flag = 1;
   }
 
