@@ -10,22 +10,13 @@
 |*		Han description table					*|
 \*======================================================================*/
 
-struct hanentry hantable [] =
-{
+struct raw_hanentry hantable [] = {
 #ifdef use_cjk_tables
-
-#ifndef __EMX__	/* EMX cannot compile this ("virtual memory exhausted") */
-
 #include "handescr.t"
-
-#endif
-
 #else
-
-# ifdef __TURBOC__
+#if defined(__TURBOC__) || defined(vms)
 	{-1}
-# endif
-
+#endif
 #endif
 };
 
