@@ -3,6 +3,11 @@
 |*		keyboard input						*|
 \*======================================================================*/
 
+#ifdef vms
+#define NOSTRINGTABLES
+#endif
+
+
 #include "mined.h"
 #include "io.h"
 #include "termprop.h"
@@ -67,7 +72,7 @@ int mouse_xpos, mouse_ypos, mouse_prevxpos, mouse_prevypos;
 int mouse_shift = 0;		/* modifier status */
 FLAG report_release = False;	/* selectively suppress release event */
 FLAG window_focus = True;	/* does the window have mouse/keyboard focus? */
-static int mouse_buttons_pressed = 0;	/* count pressed buttons */
+int mouse_buttons_pressed = 0;	/* count pressed buttons */
 
 
 /*======================================================================*\
@@ -104,7 +109,9 @@ extern struct fkeyentry fkeymap_scoansi [];
 extern struct fkeyentry fkeymap_interix [];
 
 static struct fkeyentry * fkeymap_spec = fkeymap_xterm;
+#ifndef vms
 extern struct fkeyentry fkeymap_terminfo [];
+#endif
 
 void
 set_fkeymap (term)
@@ -152,7 +159,7 @@ set_fkeymap (term)
 
 
 char * keyboard_mapping = "--";
-static char * last_keyboard_mapping = "--";
+char * last_keyboard_mapping = "--";
 static char * next_keyboard_mapping = "";
 
 static keymaptabletype keyboard_map = (keymaptabletype) NIL_PTR;
@@ -234,14 +241,13 @@ FLAG
 selectKEYMAP (script)
   char * script;
 {
-  int i;
-
   if (script == NIL_PTR || * script == '\0') {
 	set_keymap ((keymaptabletype) NIL_PTR, "--", keymappingtable [0].shortcut);
 	flags_changed = True;
 	return True;
   } else {
 	/* map alternative shortcuts */
+	int i;
 	if (strisprefix ("el", script)) {
 		script = "gr";
 	} else if (strisprefix ("ru", script)) {
@@ -265,23 +271,24 @@ selectKEYMAP (script)
 /**
    Set keyboard mapping.
  */
-void
+FLAG
 setKEYMAP (script)
   char * script;
 {
   if (script != NIL_PTR && * script == '-') {
 	/* set keymap as secondary map */
 	script ++;
-	selectKEYMAP (script);
-	selectKEYMAP (NIL_PTR);
+	(void) selectKEYMAP (script);
+	(void) selectKEYMAP (NIL_PTR);
   } else if (script != NIL_PTR) {
 	char * secondary = strchr (script, '-');
 	if (secondary != NIL_PTR) {
 		secondary ++;
-		selectKEYMAP (secondary);
+		(void) selectKEYMAP (secondary);
 	}
-	selectKEYMAP (script);
+	return selectKEYMAP (script);
   }
+  return True;
 }
 
 #ifdef not_used
@@ -291,7 +298,7 @@ setKEYMAP (script)
 void
 cycleKEYMAP ()
 {
-  selectKEYMAP (next_keyboard_mapping);
+  (void) selectKEYMAP (next_keyboard_mapping);
 }
 #endif
 
@@ -305,7 +312,7 @@ setupKEYMAP ()
   if (allow_keymap) {
 	if (hop_flag > 0) {
 		hop_flag = 0;
-		selectKEYMAP (next_keyboard_mapping);
+		(void) selectKEYMAP (next_keyboard_mapping);
 	} else {
 		handleKeymapmenu ();
 	}
@@ -383,6 +390,26 @@ get_digits (result)
 |*				Terminal input				*|
 \*======================================================================*/
 
+#ifdef debug_keytrack
+
+void
+do_trace_keytrack (tag, c)
+  char * tag;
+  unsigned long c;
+{
+  printf ("[%s] ", tag);
+  if (command (c) == COMPOSE) {
+	printf ("COMPOSE");
+  } else if (c == FUNcmd) {
+	printf ("FUNcmd");
+  } else {
+	printf ("%02X", c);
+  }
+  printf (" [keyshift %X]\n", keyshift);
+}
+
+#endif
+
 #define dont_debug_mapped_keyboard
 
 #define max_stamps 10
@@ -404,6 +431,7 @@ void
 timestamp_readchar ()
 {
 #ifndef __TURBOC__
+#ifndef VAXC
   struct timeval now;
   long sec;
   long msec;
@@ -435,6 +463,7 @@ timestamp_readchar ()
 
   average_delta_readchar = total_deltas / n_stamps;
 #endif
+#endif
 }
 
 /*
@@ -461,7 +490,7 @@ readchar ()
 
   if (winchg && waitingforinput == False) {
 	/* In the Unix version, this is now done in __readchar () */
-	RDwin ();
+	RDwinchg ();
   }
   /* must save waitingforinput flag since in the MSDOS version, 
      readchar can be called recursively */
@@ -470,14 +499,15 @@ readchar ()
   waitingforinput = True;
 
   c = _readchar ();
+  trace_keytrack ("readchar:_readchar", c);
   /* check report_release to suppress subsequent release actions 
      (esp after menu item selection) 
      when multiple buttons are pressed simultaneously */
   while (report_release == False &&
-	 command (c) == MOUSEescape && mouse_button == releasebutton)
+	 command (c) == MOUSEfunction && mouse_button == releasebutton)
   {
 #ifdef debug_mouse
-	printf ("readchar: ignoring MOUSE release (report_release == False)\n");
+	printf ("readchar: ignoring mouse release (report_release == False)\n");
 #endif
 	c = _readchar ();
   }
@@ -525,30 +555,6 @@ readchar_nokeymap ()
   return c;
 }
 
-/*
-   readcharacter_allbuttons reads like readcharacter
-   but always reports mouse button release
-   and also reports window size change as function RDwin
- */
-unsigned long
-readcharacter_allbuttons ()
-{
-  unsigned long c;
-  int prev_report_winchg = report_winchg;
-
-  report_winchg = 1;
-
-#ifdef debug_mouse
-  printf ("readchar_allbuttons: report_release = True\n");
-#endif
-  report_release = True;
-  c = readcharacter ();
-
-  report_winchg = prev_report_winchg;
-
-  return c;
-}
-
 /*
    readcharacter_mapping () reads one character
    either with or without keyboard mapping
@@ -561,12 +567,12 @@ readcharacter_mapping (map_keyboard, return_unicode)
 {
   int utfcount = 1;
   unsigned long unichar;
-  int byte2;
 
   if (map_keyboard) {
 	unichar = readchar ();
   } else {
 	unichar = readchar_nokeymap ();
+	trace_keytrack ("readcharacter:readchar_nokeymap", unichar);
 	/* double check needed here because command ('\n') == SNL */
 	if (unichar == FUNcmd && command (unichar) == SNL) {
 		return '\r';
@@ -596,16 +602,17 @@ readcharacter_mapping (map_keyboard, return_unicode)
 #endif
 		FOCUSin ();
 		skipmouse = True;
-	} else if (command (unichar) == MOUSEescape) {
+		return FUNcmd;
+	} else if (command (unichar) == MOUSEfunction) {
 		if (mouse_button == releasebutton && skip_1release) {
 #ifdef debug_mouse
-			printf ("skipping 1 MOUSE release event\n");
+			printf ("skipping 1 mouse release event\n");
 #endif
 			skipmouse = True;
 			skip_1release = False;
 		} else if (mouse_button != releasebutton && skip_1click) {
 #ifdef debug_mouse
-			printf ("skipping 1 MOUSE event\n");
+			printf ("skipping 1 mouse event\n");
 #endif
 			skipmouse = True;
 			skip_1click = False;
@@ -613,7 +620,7 @@ readcharacter_mapping (map_keyboard, return_unicode)
 		} else if (! window_focus) {
 			/* should not be needed anymore */
 #ifdef debug_mouse
-			printf ("skipping MOUSE event while focus out\n");
+			printf ("skipping mouse event while focus out\n");
 #endif
 			skipmouse = True;
 		}
@@ -624,11 +631,11 @@ readcharacter_mapping (map_keyboard, return_unicode)
 	} else {
 		/* WINCH exception isn't passing by here... */
 #ifdef debug_mouse
-		if (command (unichar) == MOUSEescape) {
+		if (command (unichar) == MOUSEfunction) {
 			if (mouse_button == releasebutton) {
-				printf ("passing MOUSE release event\n");
+				printf ("passing mouse release event\n");
 			} else {
-				printf ("passing MOUSE event\n");
+				printf ("passing mouse event\n");
 			}
 		}
 #endif
@@ -677,7 +684,7 @@ readcharacter_mapping (map_keyboard, return_unicode)
 			| (readchar_nokeymap () << 8)
 			| readchar_nokeymap ();
 	} else {
-		byte2 = readchar_nokeymap ();
+		int byte2 = readchar_nokeymap ();
 		if (text_encoding_tag == 'G' && '0' <= byte2 && byte2 <= '9') {
 			unichar = (unichar << 24)
 				| (byte2 << 16)
@@ -766,6 +773,31 @@ readcharacter_unicode_mapped ()
   return readcharacter_mapping (True, True);
 }
 
+/*
+   readcharacter_allbuttons always reports mouse button release
+   and also reports window size change as function RDwin
+ */
+unsigned long
+readcharacter_allbuttons (map_keyboard)
+  FLAG map_keyboard;
+{
+  unsigned long c;
+  int prev_report_winchg = report_winchg;
+
+  report_winchg = 1;
+
+#ifdef debug_mouse
+  printf ("readchar_allbuttons: report_release = True\n");
+#endif
+  report_release = True;
+
+  c = readcharacter_mapping (map_keyboard, False);
+
+  report_winchg = prev_report_winchg;
+
+  return c;
+}
+
 
 /*-------------------------------------------------------------------------*/
 
@@ -785,6 +817,7 @@ do_trace_mouse (tag)
 	"wheeldown"};
 
   printf ("[%s] ", tag);
+  printf ("%X ", mouse_shift);
   if (mouse_button != releasebutton) {
 	printf ("%s", button_name [mouse_button]);
   } else {
@@ -826,14 +859,50 @@ fix_mouse_release_event ()
 /*
  * DIRECT...getxy () reads in the information part of a cursor escape sequence
  */
+static
+int
+DIRECTxtermgetbut ()
+{
+  int but = _readchar_nokeymap ();
+  if (but == quit_char) {
+	quit = True;
+	return 0;
+  }
+  if (use_mouse_extended) {
+	if ((but & 0xE0) == 0xC0) {
+		but &= 0x1F;
+		but = (but << 6) | (_readchar_nokeymap () & 0x3F);
+	}
+  }
+  return but;
+}
+
+static
+int
+DIRECTxtermgetpos ()
+{
+  int pos = _readchar_nokeymap ();
+  if (pos == quit_char) {
+	quit = True;
+	return 0;
+  }
+  if (use_mouse_extended) {
+	if ((pos & 0xE0) == 0xC0) {
+		pos &= 0x1F;
+		pos = (pos << 6) | (_readchar_nokeymap () & 0x3F);
+	}
+  } else if (pos == '\0') {
+	pos = 256;
+  }
+  return pos - 33;
+}
+
 static
 void
 DIRECTxtermgetxy (code)
   char code;	/* 'M' for normal mouse report, 't'/'T' for tracking */
 {
   character button;
-  int xpos;
-  int ypos;
 
   notice_previous_click ();
 
@@ -847,10 +916,15 @@ DIRECTxtermgetxy (code)
 	button = _readchar_nokeymap ();
 	mouse_button = releasebutton;
   } else {	/* code == 'M' */
-	button = _readchar_nokeymap ();
-	if (button == quit_char) {
-		quit = True;
+	if (xterm_version >= 268) {
+		button = DIRECTxtermgetbut ();
+	} else {
+		button = _readchar_nokeymap ();
+		if (button == quit_char) {
+			quit = True;
+		}
 	}
+
 	if (bidi_screen && button == '$') {
 		/* fix mlterm quirk for mouse wheel scroll down */
 		button = 'a';
@@ -883,16 +957,8 @@ DIRECTxtermgetxy (code)
 
   fix_mouse_release_event ();
 
-  xpos = _readchar_nokeymap ();
-  if (xpos == quit_char) {
-	quit = True;
-  }
-  mouse_xpos = xpos - 33;
-  ypos = _readchar_nokeymap ();
-  if (ypos == quit_char) {
-	quit = True;
-  }
-  mouse_ypos = ypos - 33 - MENU;
+  mouse_xpos = DIRECTxtermgetpos ();
+  mouse_ypos = DIRECTxtermgetpos () - MENU;
 
 #ifdef debug_mouse
   printf ("mouse %c: ", code);
@@ -973,7 +1039,7 @@ DIRECTxterm ()
   printf ("DIRECTxterm\n");
 #endif
   DIRECTxtermgetxy ('M');
-  MOUSEescape ();
+  MOUSEfunction ();
 }
 
 void
@@ -983,7 +1049,7 @@ TRACKxterm ()
   printf ("TRACKxterm\n");
 #endif
   DIRECTxtermgetxy ('t');
-  MOUSEescape ();
+  MOUSEfunction ();
 }
 
 void
@@ -993,14 +1059,14 @@ TRACKxtermT ()
   printf ("TRACKxtermT\n");
 #endif
   DIRECTxtermgetxy ('T');
-  MOUSEescape ();
+  MOUSEfunction ();
 }
 
 void
 DIRECTcrttool ()
 {
   DIRECTcrttoolgetxy ();
-  MOUSEescape ();
+  MOUSEfunction ();
 }
 
 
@@ -1115,8 +1181,6 @@ CURSgetxy ()
 
 /*-------------------------------------------------------------------------*/
 
-#ifndef msdos
-
 /* max. length of function key sequence to be detected
    (depending on the fkeymap table, approx. 7),
    or of keyboard mapping sequence plus potential mapping rest
@@ -1227,129 +1291,6 @@ q_get ()
   return c;
 }
 
-/*
- * Look up key sequence in fkeymap table.
- * if matchmode == 0:
- * findkey (str) >=  0: escape sequence found
- *				(str == fkeymap [findkey (str)].fk)
- *			keyproc = fkeymap [i].fp (side effect)
- *		 == -1: escape sequence prefix recognised
- *				(str is prefix of some entry in fkeymap)
- *		 == -2: escape sequence not found
- *				(str is not contained in fkeymap)
- * * found == index of exact match if found
- * if matchmode == 1 (not used):
- * return -1 if any prefix detected (even if other exact match found)
- * if matchmode == 2 (not used):
- * don't return -1, only report exact or no match
- */
-static
-int
-findkeyin (str, fkeymap, matchmode, found)
-  char * str;
-  struct fkeyentry * fkeymap;
-  int matchmode;
-  int * found;
-{
-  int lastmatch = 0;	/* last index with string matching prefix */
-  register int i;
-  int ret = -2;
-
-  * found = -1;
-
-  if (fkeymap [0].fk == NIL_PTR) {
-	return ret;
-  }
-
-  i = lastmatch;
-  do {
-	char * strpoi = str;
-	char * mappoi = fkeymap [i].fk;
-	do {
-		/* compare str with map entry */
-		if (* strpoi != * mappoi) {
-			break;
-		}
-		if (* strpoi) {
-			strpoi ++;
-			mappoi ++;
-		}
-	} while (* strpoi);
-
-/*	if (strncmp (str, fkeymap [i].fk, strlen (str)) == 0) {*/
-	if (* strpoi == '\0') {
-		/* str is prefix of current entry */
-		lastmatch = i;
-/*		if (strlen (str) == strlen (fkeymap [i].fk)) {*/
-		if (* mappoi == '\0') {
-			/* str is equal to current entry */
-			* found = i;
-			keyproc = fkeymap [i].fp;
-			if (! keyproc) {
-				keyproc = I;
-			}
-			fkeyshift = fkeymap [i].fkeyshift;
-			if (matchmode == 1) {
-				/* return index unless other prefix 
-				   was or will be found */
-				if (ret == -2) {
-					ret = i;
-				}
-			} else {
-				return i;
-			}
-		} else {
-			/* str is partial prefix of current entry */
-			if (matchmode == 1) {
-				/* return -1 but continue to search 
-				   for exact match */
-				ret = -1;
-			} else if (matchmode != 2) {
-				return -1;
-			} else {
-			}
-		}
-	}
-	++ i;
-	if (fkeymap [i].fk == NIL_PTR) {
-		i = 0;
-		return ret;
-	}
-  } while (i != lastmatch);
-
-  return ret;
-}
-
-static
-int
-findkey (str)
-  char * str;
-{
-  int dummy;
-  int res;
-
-  res = findkeyin (str, fkeymap_spec, 0, & dummy);
-#ifdef debug_findkey
-  printf ("<^[%s> _spec %d\n", str + 1, res);
-#endif
-  if (res != -2) {
-	return res;
-  }
-
-  res = findkeyin (str, fkeymap_terminfo, 0, & dummy);
-#ifdef debug_findkey
-  printf ("<^[%s> _terminfo %d\n", str + 1, res);
-#endif
-  if (res != -2) {
-	return res;
-  }
-
-  res = findkeyin (str, fkeymap, 0, & dummy);
-#ifdef debug_findkey
-  printf ("<^[%s> _ %d\n", str + 1, res);
-#endif
-  return res;
-}
 
 /*
    Look up key sequence in keyboard mapping table.
@@ -1488,10 +1429,146 @@ mapkb (str, kbmap, matchmode, found, mapped)
 #endif
 }
 
-#endif	/* ifndef msdos */
+
+#ifndef msdos
+
+/*
+ * Look up key sequence in fkeymap table.
+ * if matchmode == 0:
+ * findkey (str) >=  0: escape sequence found
+ *				(str == fkeymap [findkey (str)].fk)
+ *			keyproc = fkeymap [i].fp (side effect)
+ *		 == -1: escape sequence prefix recognised
+ *				(str is prefix of some entry in fkeymap)
+ *		 == -2: escape sequence not found
+ *				(str is not contained in fkeymap)
+ * * found == index of exact match if found
+ * if matchmode == 1 (not used):
+ * return -1 if any prefix detected (even if other exact match found)
+ * if matchmode == 2 (not used):
+ * don't return -1, only report exact or no match
+ */
+static
+int
+findkeyin (str, fkeymap, matchmode, found)
+  char * str;
+  struct fkeyentry * fkeymap;
+  int matchmode;
+  int * found;
+{
+  int lastmatch = 0;	/* last index with string matching prefix */
+  register int i;
+  int ret = -2;
+
+  * found = -1;
+
+  if (fkeymap [0].fk == NIL_PTR) {
+	return ret;
+  }
+
+  i = lastmatch;
+  do {
+	char * strpoi = str;
+	char * mappoi = fkeymap [i].fk;
+	do {
+		/* compare str with map entry */
+		if (* strpoi != * mappoi) {
+			break;
+		}
+		if (* strpoi) {
+			strpoi ++;
+			mappoi ++;
+		}
+	} while (* strpoi);
+
+/*	if (strncmp (str, fkeymap [i].fk, strlen (str)) == 0) {*/
+	if (* strpoi == '\0') {
+		/* str is prefix of current entry */
+		lastmatch = i;
+/*		if (strlen (str) == strlen (fkeymap [i].fk)) {*/
+		if (* mappoi == '\0') {
+			/* str is equal to current entry */
+			* found = i;
+			keyproc = fkeymap [i].fp;
+			if (! keyproc) {
+				keyproc = I;
+			}
+			fkeyshift = fkeymap [i].fkeyshift;
+			if (matchmode == 1) {
+				/* return index unless other prefix 
+				   was or will be found */
+				if (ret == -2) {
+					ret = i;
+				}
+			} else {
+				return i;
+			}
+		} else {
+			/* str is partial prefix of current entry */
+			if (matchmode == 1) {
+				/* return -1 but continue to search 
+				   for exact match */
+				ret = -1;
+			} else if (matchmode != 2) {
+				return -1;
+			} else {
+			}
+		}
+	}
+	++ i;
+	if (fkeymap [i].fk == NIL_PTR) {
+		i = 0;
+		return ret;
+	}
+  } while (i != lastmatch);
+
+  return ret;
+}
+
+static
+int
+findkey (str)
+  char * str;
+{
+  int dummy;
+  int res;
+
+  res = findkeyin (str, fkeymap_spec, 0, & dummy);
+#ifdef debug_findkey
+  printf ("<^[%s> _specific [%d]\n", str + 1, res);
+#endif
+
+  if (res != -2) {
+	return res;
+  }
+
+  res = findkeyin (str, fkeymap, 0, & dummy);
+#ifdef debug_findkey
+  printf ("<^[%s> _deftable [%d]\n", str + 1, res);
+#endif
+
+  if (res != -2) {
+	return res;
+  }
+
+#ifndef vms
+  res = findkeyin (str, fkeymap_terminfo, 0, & dummy);
+# ifdef debug_findkey
+  printf ("<^[%s> _terminfo [%d]\n", str + 1, res);
+# endif
+#endif
+
+  return res;
+}
+
+#endif	/* #ifndef msdos */
 
 
 #ifdef CURSES
+# ifdef vms
+typedef unsigned int chtype;
+# endif
+
 chtype key;
 
 extern int curs_readchar _((void));
@@ -1514,13 +1591,18 @@ curs_readchar ()
 {
   unsigned long keycode;
   keycode = __readchar ();
+#ifdef vms
+  if (keycode >= 0x100) {
+	printf ("readchar: vms curses key %03X\n", keycode); sleep (1);
+  }
+#else
   while (keycode >= KEY_MIN) {
 	switch (keycode) {
 #ifdef KEY_RESIZE
 	case KEY_RESIZE:
-  printf ("curses resize key received \n"); sleep (1);
-			keyproc = showfkey;
-			return FUNcmd;
+		printf ("curses resize key received \n"); sleep (1);
+		keyproc = showfkey;
+		return FUNcmd;
 #endif
 #ifdef KEY_SHIFT_L
 	case KEY_SHIFT_L:
@@ -1529,12 +1611,14 @@ curs_readchar ()
 	case KEY_CONTROL_R:
 	case KEY_ALT_L:
 	case KEY_ALT_R:
-			break;
+		break;
 #endif
-	default:	return keycode;
+	default:
+		return keycode;
 	}
 	keycode = __readchar ();
   }
+#endif
 
   return keycode;
 }
@@ -1546,8 +1630,8 @@ struct {
 	chtype keycode;
 	voidfunc func;
 	unsigned char fkeyshift;
-} cursfunc [] =
-{
+} cursfunc [] = {
+#ifndef vms
 	{KEY_F(1), F1},
 	{KEY_F(2), F2},
 	{KEY_F(3), F3},
@@ -1570,6 +1654,12 @@ struct {
 	{KEY_DOWN, MDN},
 	{KEY_LEFT, MLF},
 	{KEY_RIGHT, MRT},
+	{KEY_B2, HOP},
+	{KEY_A1, HOMEkey},
+	{KEY_A3, PU},
+	{KEY_C1, ENDkey},
+	{KEY_C3, PD},
+#endif
 #ifdef KEY_A2
 	{KEY_A2, MUP},
 #endif
@@ -1582,11 +1672,6 @@ struct {
 #ifdef KEY_B3
 	{KEY_B3, MRT},
 #endif
-	{KEY_B2, HOP},
-	{KEY_A1, HOMEkey},
-	{KEY_A3, PU},
-	{KEY_C1, ENDkey},
-	{KEY_C3, PD},
 	{0x1BF, BLINE},
 	{0x1C0, ELINE},
 
@@ -1658,7 +1743,7 @@ _readchar_nokeymap ()
   return c;
 }
 
-#ifndef msdos
+
 static int choose_char _((char *));
 
 static character rest_queue [MAXCODELEN + 1];
@@ -1691,7 +1776,6 @@ putback_rest (rest)
 	}
 }
 
-#endif
 
 /* read 1 byte from queue; for curses, however, function 
    key indications must be passed on which are longer than 8 bits
@@ -1699,12 +1783,9 @@ putback_rest (rest)
 int
 read1byte ()
 {
-#ifndef msdos
-  character ch;
-  character * cr;
   if (have_rest_queue) {
-	ch = * rest_queue;
-	cr = rest_queue;
+	character * cr = rest_queue;
+	character ch = * rest_queue;
 	while (* (cr + 1) != '\0') {
 		* cr = * (cr + 1);
 		cr ++;
@@ -1715,7 +1796,6 @@ read1byte ()
 	}
 	return ch;
   } else
-#endif
   {
 #ifdef CURSES
 	return curs_readchar ();
@@ -1737,8 +1817,6 @@ read1byte ()
 #define trace_terminal_response(what)
 #endif
 
-#ifndef msdos
-
 static
 int
 bytereadyafter (msec, debug_tag)
@@ -1749,7 +1827,7 @@ bytereadyafter (msec, debug_tag)
 	trace_terminal_response ("from rest queue");
 	return 1;
   } else {
-	if (inputreadyafter (input_fd, msec) > 0) {
+	if (inputreadyafter (msec) > 0) {
 		trace_terminal_response ("response");
 		return 1;
 	} else {
@@ -1775,24 +1853,6 @@ char_ready_within (msec, debug_tag)
   }
 }
 
-#else	/* ifndef msdos */
-
-int
-char_ready_within (msec, debug_tag)
-  int msec;
-  char * debug_tag;
-{
-  if (inputreadyafter (input_fd, msec) > 0) {
-	trace_terminal_response ("response");
-	return 1;
-  } else {
-	trace_terminal_response ("timeout");
-	return 0;
-  }
-}
-
-#endif
-
 
 #ifndef msdos
 #define radical_stroke
@@ -1800,8 +1860,7 @@ char_ready_within (msec, debug_tag)
 
 
 #ifdef radical_stroke
-static menuitemtype radicalsmenu [] =
-{
+static menuitemtype radicalsmenu [] = {
 	/* 1 stroke radicals */
 	{"①.1:一 2:丨 3:丶 4:丿乀乁 5:乙乚乛 6:亅", dummyfunc, ""},
 	/* 2 strokes radicals */
@@ -1865,25 +1924,24 @@ unsigned long
 _readchar ()
 {
   unsigned long ch;
-#ifndef msdos
+  char * choice;
+  int choice_index;
+  /* queue handling */
   int res;
   char * found;
   char * mapped;
   char * prev_found = NIL_PTR;
-#endif
-  char * choice;
-  int choice_index;
 
 
   if (escape_delay == 0 || keymap_delay == 0) {
-	char * env = getenv ("ESCDELAY");
+	char * env = envvar ("ESCDELAY");
 	if (env) {
 		(void) scan_int (env, & escape_delay);
 	}
 	if (escape_delay == 0) {
 		escape_delay = default_escape_delay;
 	}
-	env = getenv ("MAPDELAY");
+	env = envvar ("MAPDELAY");
 	if (env) {
 		(void) scan_int (env, & keymap_delay);
 	}
@@ -1893,12 +1951,10 @@ _readchar ()
   }
 
 
-#ifndef msdos
   if (q_len () > 0) {
 	return q_get ();
   }
   queue_mode_mapped = False;
-#endif
 
   keyshift = 0;
 
@@ -1978,27 +2034,33 @@ _readchar ()
 #ifdef CURSES
 
   key = read1byte ();
+#ifdef vms
+  if (key >= 0x100) {
+	printf ("_readchar: vms curses key %03X\n", key); sleep (1);
+  }
+#else
   if (key >= KEY_MIN) {
 #ifdef KEY_MOUSE
 	if (key == KEY_MOUSE) {
 		CURSgetxy ();
-		keyproc = MOUSEescape;
+		keyproc = MOUSEfunction;
 		return FUNcmd;
 	}
 #endif
 	keyproc = lookup_curskey (key);
 	return FUNcmd;
   }
+#endif
   ch = key;
 
-#else	/* ifdef CURSES */
+#else	/* #ifdef CURSES */
 
 #ifdef dosmouse
   {
 	int ich = __readchar ();
-	if (ich < 0) {
+	if (ich == FUNcmd) {
 		DIRECTdosmousegetxy ();
-		keyproc = MOUSEescape;
+		keyproc = MOUSEfunction;
 		return FUNcmd;
 	} else {
 		ch = ich;
@@ -2012,7 +2074,7 @@ _readchar ()
   }
 #endif
 
-#endif	/* else CURSES */
+#endif	/* #else CURSES */
 
 
 #ifdef msdos
@@ -2032,11 +2094,11 @@ _readchar ()
 # else
 	ch = read1byte ();
 # endif
-	keyproc = pc_fkey_map [ch].fp;
+	keyproc = pc_xkey_map [ch].fp;
 	if (! keyproc) {
 		keyproc = I;
 	}
-	keyshift = pc_fkey_map [ch].fkeyshift;
+	keyshift = pc_xkey_map [ch].fkeyshift;
 # ifdef debug_whatever
 	if ((voidfunc) keyproc == (voidfunc) I) {
 		return ch;
@@ -2046,11 +2108,12 @@ _readchar ()
 # endif
 	return FUNcmd;
   } else {
-	return ch;
+	q_put (ch);
   }
 
-#else
+#else	/* #ifdef msdos */
 
+  trace_keytrack ("_readchar<033", ch);
   if (ch == '\033') {
 	char second_esc_char = '\0';
 
@@ -2159,6 +2222,10 @@ check_ctrl_byte:
 					/* ESC O11P → ESC OP + alt etc */
 					/* HP: ESC 11p → ESC p + alt etc */
 					keyshift |= applkeypad_mask | ((ch & ~'0') + 1);
+				} else if (ch == ';') {
+					/* mlterm: ESC O1;2A → ESC OA + shift etc */
+					ch = read1byte ();
+					keyshift |= ((ch & ~'0') - 1);
 				} else {
 					q_put ('1');
 					q_put (ch);
@@ -2230,7 +2297,7 @@ check_ctrl_byte:
 				   goto avoidance effort
 				*/
 #ifdef debug_kbesc
-				printf ("GOTO check_ctrl_byte\n");
+				printf ("goto check_ctrl_byte\n");
 #endif
 				goto check_ctrl_byte;
 			}
@@ -2438,27 +2505,30 @@ check_ctrl_byte:
 
 		keyshift |= fkeyshift;
 
-		if (keyproc == DIRECTxterm) {
+		if (keyproc == ESCAPE) {
+			/* mintty application escape key mode */
+			return '\033';
+		} else if (keyproc == DIRECTxterm) {
 #ifdef debug_mouse
 			printf ("scanned xterm mouse escape\n");
 #endif
 			DIRECTxtermgetxy ('M');
-			keyproc = MOUSEescape;
+			keyproc = MOUSEfunction;
 		} else if (keyproc == TRACKxterm) {
 #ifdef debug_mouse
 			printf ("scanned xterm mouse tracking escape\n");
 #endif
 			DIRECTxtermgetxy ('t');
-			keyproc = MOUSEescape;
+			keyproc = MOUSEfunction;
 		} else if (keyproc == TRACKxtermT) {
 #ifdef debug_mouse
 			printf ("scanned xterm mouse tracking escape\n");
 #endif
 			DIRECTxtermgetxy ('T');
-			keyproc = MOUSEescape;
+			keyproc = MOUSEfunction;
 		} else if (keyproc == DIRECTcrttool) {
 			DIRECTcrttoolgetxy ();
-			keyproc = MOUSEescape;
+			keyproc = MOUSEfunction;
 		}
 
 		return FUNcmd;
@@ -2467,8 +2537,11 @@ check_ctrl_byte:
 	q_put (ch);
   }
 
+#endif	/* #else msdos */
+
   if (allow_keymap && dont_suppress_keymap
 		&& keyboard_map != (keymaptabletype) NIL_PTR) {
+	disable_modify_keys (True);
 	while  ((res = mapkb (ctrl_queue, keyboard_map, 1, & found, & mapped)) == -1
 		 /* prefix of some table entry */
 		&& q_notfull ()
@@ -2481,6 +2554,7 @@ check_ctrl_byte:
 		 */
 		q_put (_readchar_nokeymap ());
 		plainly like this, we get an endless loop...
+		now using disable_modify_keys workaround
 #else
 		q_put (read1byte ());
 #endif
@@ -2489,6 +2563,7 @@ check_ctrl_byte:
 			prev_found = found;
 		}
 	}
+	disable_modify_keys (False);
 	/* possible results:
 		res >= 0: match found, no further prefix match
 		res == -1, prev_found != NIL_PTR: prefix with previous match
@@ -2578,12 +2653,9 @@ handle_mapped:
 	return q_get ();
   }
 
-#endif	/* #ifdef msdos #else */
 }
 
 
-#ifndef msdos
-
 #define dont_debug_choose_char
 
 static
@@ -2684,7 +2756,7 @@ choose_char (choices)
 		printf ("choiceline <%s>\n", thischoiceline);
 #endif
 		/* fill menu item with collected data */
-		fill_menuitem (& choicemenu [li], thischoiceline);
+		fill_menuitem (& choicemenu [li], thischoiceline, NIL_PTR);
 
 		thischoiceline = choicelinepoi;
 	}
@@ -2728,8 +2800,6 @@ choose_char (choices)
 	}
 }
 
-#endif
-
 
 /*======================================================================*\
 |*				End					*|
