@@ -15,224 +15,81 @@
 
 #define Sch (voidfunc) Scharacter
 
-voidfunc key_map [256] =
-{	/* characters to functions map */
+voidfunc key_map [256] = {
+	/* characters to functions map */
   /* ^@ ^A ^B ^C ^D ^E ^F ^G */
-		MARK, MPW, DCC, PD, MRT, MUP, MNW, GOTO,
+		MARK, MPW, DCC, PD, MRT, MUP, MNW, GOHOP,
   /* ^H ^I ^J ^K ^L ^M ^N ^O */
 		DPC, Sch, Sch, DLN, HOP, SNL, RS, LIB,
   /* ^P ^Q ^R ^S ^T ^U ^V ^W */
 		PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
   /* ^X ^Y ^Z ^[ ^\ ^] ^^ ^_ */
 		MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
-  /* 040-057 */	Sch, Sch, Sdoublequote, Sch, Sch, Sch, Sch, Ssinglequote,
+  /* 20-2F 040-057 */	Sch, Sch, Sdoublequote, Sch, Sch, Sch, Sch, Ssinglequote,
 		Sch, Sch, Sch, Sch, Sch, Sdash, Sch, Sch,
-  /* 060-077 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 100-117 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 120-137 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 140-157 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 160-177 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, 
+  /* 30-3F 060-077 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 40-4F 100-117 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 50-5F 120-137 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 60-6F 140-157 */	Sgrave, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 70-7F 160-177 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, 
 		Sch, Sch, Sch, Sch, Sch, Sch, Sch, DELchar,
-  /* 200-217 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 220-237 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 240-257 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 260-277 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 300-317 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 320-337 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 340-357 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
-  /* 360-377 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 80-8F 200-217 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* 90-9F 220-237 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* A0-AF 240-257 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* B0-BF 260-277 */	Sch, Sch, Sch, Sch, Sacute, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* C0-CF 300-317 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* D0-DF 320-337 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* E0-EF 340-357 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+  /* F0-FF 360-377 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
 };
 
-voidfunc ws_key_map [32] =
-{	/* characters to functions map */
+voidfunc mined_key_map [32] = {
+	/* control characters to functions map */
+  /* ^@-^G */	MARK, MPW, DCC, PD, MRT, MUP, MNW, GOHOP,
+  /* ^H-^O */	DPC, Sch, Sch, DLN, HOP, SNL, RS, LIB,
+  /* ^P-^W */	PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
+  /* ^X-^_ */	MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
+};
+
+voidfunc ws_key_map [32] = {
+	/* control characters to functions map */
   /* ^@-^G */	I, MPW, JUS, PD, MRT, MUP, MNW, DCC,
   /* ^H-^O */	MLF, Sch, Sch, ctrlK, RS, SNL, LIB, ctrlO,
   /* ^P-^W */	CTRLINS, ctrlQ, PU, MLF, DNW, (voidfunc) BADch, TOGINS, SU,
   /* ^X-^_ */	MDN, DLINE, SD, ESCAPE, CANCEL, (voidfunc) BADch, (voidfunc) BADch, (voidfunc) BADch,
 };
 
-voidfunc pico_key_map [32] =
-{	/* characters to functions map */
+voidfunc pico_key_map [32] = {
+	/* control characters to functions map */
   /* ^@-^G */	MNW, GO_LINEBEGIN , MLF, FS, DCC, GO_LINEEND , MRT, HELP,
   /* ^H-^O */	DPC, Sch, JUS, DLN, RDwin, SNL, MDN, WTU,
   /* ^P-^W */	MUP, I, INSFILE, I, SPELL, PASTE, PD, SFW,
   /* ^X-^_ */	QUED, PU, I, ESCAPE, CANCEL, MARK, toggleMARK, CTRLINS,
 };
 
-voidfunc emacs_key_map [32] =
-{	/* characters to functions map */
+voidfunc emacs_key_map [32] = {
+	/* control characters to functions map */
   /* ^@-^G */	MARK, GO_LINEBEGIN , MLF, I, DCC, GO_LINEEND , MRT, CANCEL,
   /* ^H-^O */	HELP, Sch, Sch, DLN, RD, SNL, MDN, LIB,
   /* ^P-^W */	MUP, CTRLINS, SRV, SFW, I, (voidfunc) REPT, PD, CUT,
   /* ^X-^_ */	EMAX, PASTE, SUSP, META, HOP, I, HOP, UNDO,
 };
 
-
-/* mapping DOS 2-byte (leading 0) function key codes -> functions
-   function keys deliver NUL and the following codes:
-block1:	RGISOQHKPM
-block2:	GHIKMOPQRS	Pos1 Up PgUp Left Center Right End Down PdDn Ins
-ctrl:	st <-/->
-ctrl:	wu	Pos1 End
-	0x84v	PgUp PgDn	(first char is a diaeresis in CP437)
-F1-F10:	;<=>?@ABCD
-shift:	TUVWXYZ[\]
-ctrl:	^_`abcdefg
-alt:	hijklmnopq
-*/
-struct pc_fkeyentry pc_fkey_map [256] =
-{
-   /* 00-0F */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* 10 alt-Q */ {I},
-   /* 11 alt-W */ {I},
-   /* 12 alt-E */ {EDITMENU},
-   /* 13 alt-R */ {I},
-   /* 14 alt-T */ {I},
-   /* 15 alt-Y */ {I},
-   /* 16 alt-U */ {I},
-   /* 17 alt-I */ {I},
-   /* 18 alt-O */ {OPTIONSMENU},
-   /* 19 alt-P */ {PARAMENU},
-   /* 1A-1D */ {I}, {I}, {I}, {I},
-   /* 1E alt-A */ {I},
-   /* 1F alt-S */ {SEARCHMENU},
-   /* 20 alt-D */ {I},
-   /* 21 alt-F */ {FILEMENU},
-   /* 22 alt-G */ {I},
-   /* 23 alt-H */ {HTML},
-   /* 24 alt-J */ {I},
-   /* 25 alt-K */ {I},
-   /* 26 alt-L */ {I},
-   /* 27-2B */ {I}, {I}, {I}, {I}, {I},
-   /* 2C alt-Z */ {I},
-   /* 2D alt-X */ {I},
-   /* 2E alt-C */ {I},
-   /* 2F alt-V */ {I},
-   /* 30 alt-B */ {I},
-   /* 31 alt-N */ {I},
-   /* 32 alt-M */ {I},
-   /* 33-36 */ {I}, {I}, {I}, {I},
-   /* 37 alt-* */ {I, alt_mask},
-   /* 38 */ {I},
-   /* 39 alt-control-space */ {MARK},
-   /* 3A */ {I},
-   /* 3B F1 */ {F1},
-   /* 3C F2 */ {F2},
-   /* 3D F3 */ {F3},
-   /* 3E F4 */ {F4},
-   /* 3F F5 */ {F5},
-   /* 40 F6 */ {F6},
-   /* 41 F7 */ {F7},
-   /* 42 F8 */ {F8},
-   /* 43 F9 */ {F9},
-   /* 44 F10 */ {F10},
-   /* 45-46 */ {I}, {I},
-   /* 47 home */ {HOMEkey},
-   /* 48 up */ {MUP},
-   /* 49 pgup */ {PU},
-   /* 4A alt-- */ {screenlesslines, alt_mask},
-   /* 4B lf */ {MLF},
-   /* 4C 5 */ {HOP},
-   /* 4D rt */ {MRT},
-   /* 4E alt-+ */ {screenmorelines, alt_mask},
-   /* 4F end */ {ENDkey},
-   /* 50 dn */ {MDN},
-   /* 51 pgdn */ {PD},
-   /* 52 ins */ {INSkey},
-   /* 53 del */ {DELkey},
-   /* 54 shift-F1 */ {F1, shift_mask},
-   /* 55 shift-F2 */ {F2, shift_mask},
-   /* 56 shift-F3 */ {F3, shift_mask},
-   /* 57 shift-F4 */ {F4, shift_mask},
-   /* 58 shift-F5 */ {F5, shift_mask},
-   /* 59 shift-F6 */ {F6, shift_mask},
-   /* 5A shift-F7 */ {F7, shift_mask},
-   /* 5B shift-F8 */ {F8, shift_mask},
-   /* 5C shift-F9 */ {F9, shift_mask},
-   /* 5D shift-F10 */ {F10, shift_mask},
-   /* 5E control-F1 */ {F1, ctrl_mask},
-   /* 5F control-F2 */ {F2, ctrl_mask},
-   /* 60 control-F3 */ {F3, ctrl_mask},
-   /* 61 control-F4 */ {F4, ctrl_mask},
-   /* 62 control-F5 */ {F5, ctrl_mask},
-   /* 63 control-F6 */ {F6, ctrl_mask},
-   /* 64 control-F7 */ {F7, ctrl_mask},
-   /* 65 control-F8 */ {F8, ctrl_mask},
-   /* 66 control-F9 */ {F9, ctrl_mask},
-   /* 67 control-F10 */ {F10, ctrl_mask},
-   /* 68 alt-F1 */ {F1, alt_mask},
-   /* 69 alt-F2 */ {F2, alt_mask},
-   /* 6A alt-F3 */ {F3, alt_mask},
-   /* 6B alt-F4 */ {F4, alt_mask},
-   /* 6C alt-F5 */ {F5, alt_mask},
-   /* 6D alt-F6 */ {F6, alt_mask},
-   /* 6E alt-F7 */ {F7, alt_mask},
-   /* 6F alt-F8 */ {F8, alt_mask},
-   /* 70 alt-F9 */ {F9, alt_mask},
-   /* 71 alt-F10 */ {F10, alt_mask},
-   /* 72 */ {I},
-   /* 73 control-lf */ {MLF, ctrl_mask},
-   /* 74 control-rt */ {MRT, ctrl_mask},
-   /* 75 control-end */ {ENDkey, ctrl_mask},
-   /* 76 control-pgdn */ {PD, ctrl_mask},
-   /* 77 control-home */ {HOMEkey, ctrl_mask},
-   /* 78 alt-1 */ {key_1, alt_mask},
-   /* 79 alt-2 */ {key_2, alt_mask},
-   /* 7A alt-3 */ {key_3, alt_mask},
-   /* 7B alt-4 */ {key_4, alt_mask},
-   /* 7C alt-5 */ {key_5, alt_mask},
-   /* 7D alt-6 */ {key_6, alt_mask},
-   /* 7E alt-7 */ {key_7, alt_mask},
-   /* 7F alt-8 */ {key_8, alt_mask},
-   /* 80 alt-9 */ {key_9, alt_mask},
-   /* 81 alt-0 */ {key_0, alt_mask},
-   /* 82-83 */ {I}, {I},
-   /* 84 control-pgup */ {PU, ctrl_mask},
-   /* 85 F11 */ {F11},
-   /* 86 F12 */ {F12},
-   /* 87 shift-F11 */ {F11, shift_mask},
-   /* 88 shift-F12 */ {F12, shift_mask},
-   /* 89 control-F11 */ {F11, ctrl_mask},
-   /* 8A control-F12 */ {F12, ctrl_mask},
-   /* 8B alt-F11 */ {F11, alt_mask},
-   /* 8C alt-F12 */ {F12, alt_mask},
-   /* 8D control-up */ {MUP, ctrl_mask},
-   /* 8E control-- */ {screensmaller, ctrl_mask},
-   /* 8F control-5 */ {HOP, ctrl_mask},
-   /* 90 control-+ */ {screenbigger, ctrl_mask},
-   /* 91 control-dn */ {MDN, ctrl_mask},
-   /* 92 control-ins */ {INSkey, ctrl_mask},
-   /* 93 control-del */ {DELkey, ctrl_mask},
-   /* 94 */ {I},
-   /* 95 control-/ */ {LNCI, ctrl_mask},
-   /* 96 control-* */ {I, ctrl_mask},
-   /* 97 alt-home */ {HOMEkey, alt_mask},
-   /* 98 alt-up */ {MUP, alt_mask},
-   /* 99 alt-pgup */ {PU, alt_mask},
-   /* 9A */ {I},
-   /* 9B alt-lf */ {MLF, alt_mask},
-   /* 9C */ {I},
-   /* 9D alt-rt */ {MRT, alt_mask},
-   /* 9E */ {I},
-   /* 9F alt-end */ {ENDkey, alt_mask},
-   /* A0 alt-dn */ {MDN, alt_mask},
-   /* A1 alt-pgdn */ {PD, alt_mask},
-   /* A2 alt-ins */ {INSkey, alt_mask},
-   /* A3 alt-del */ {DELkey, alt_mask},
-   /* A4 alt-/ */ {LNSW, alt_mask},
-   /* A5-AF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* B0-BF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* C0-CF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* D0-DF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* E0-EF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
-   /* F0-FF */ {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
+voidfunc windows_key_map [32] = {
+	/* control characters to functions map */
+  /* ^@-^G */	MARK, I, I, COPY, I, I, SFW, GOHOP,
+  /* ^H-^O */	REPL, Sch, Sch, I, I, SNL, I, EDIT,
+  /* ^P-^W */	PRINT, QUED, I, WT, I, I, PASTE, I,
+  /* ^X-^_ */	CUT, I, I, ESCAPE, CANCEL, I, I, CTRLINS,
 };
 
+
 /* mapping DOS extended function key codes -> functions
    keys as delivered by djgpp function getxkey, listed in keys.h
+   first half of map is also valid for function getkey (with leading 0 byte)
+   a leading : means: special MSDOS key mapping
 */
-struct pc_fkeyentry pc_xkey_map [512] =
-{
+struct pc_fkeyentry pc_xkey_map [512] = {
 					{I},
 /* #define K_Alt_Escape           0x101 */	{EXED},
 					{I},
@@ -244,7 +101,7 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Alt_W                0x111 */	{I},
 /* #define K_Alt_E                0x112 */	{EDITMENU},
 /* #define K_Alt_R                0x113 */	{I},
-/* #define K_Alt_T                0x114 */	{I},
+/* #define K_Alt_T                0x114 */	{Stag},
 /* #define K_Alt_Y                0x115 */	{I},
 /* #define K_Alt_U                0x116 */	{I},
 /* #define K_Alt_I                0x117 */	{I},
@@ -259,9 +116,9 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Alt_D                0x120 */	{I},
 /* #define K_Alt_F                0x121 */	{FILEMENU},
 /* #define K_Alt_G                0x122 */	{I},
-/* #define K_Alt_H                0x123 */	{I},
+/* #define K_Alt_H                0x123 */	{HTML},
 /* #define K_Alt_J                0x124 */	{I},
-/* #define K_Alt_K                0x125 */	{I},
+/* #define K_Alt_K                0x125 */	{toggleKEYMAP},
 /* #define K_Alt_L                0x126 */	{I},
 /* #define K_Alt_Semicolon        0x127 */	{I},
 /* #define K_Alt_Quote            0x128 */	{I},
@@ -269,7 +126,7 @@ struct pc_fkeyentry pc_xkey_map [512] =
 					{I},
 /* #define K_Alt_Backslash        0x12b */	{I},
 /* #define K_Alt_Z                0x12c */	{I},
-/* #define K_Alt_X                0x12d */	{I},
+/* #define K_Alt_X                0x12d */	{AltX},
 /* #define K_Alt_C                0x12e */	{I},
 /* #define K_Alt_V                0x12f */	{I},
 /* #define K_Alt_B                0x130 */	{I},
@@ -280,7 +137,9 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Alt_Slash            0x135 */	{I},
 					{I},
 /* #define K_Alt_KPStar           0x137 */	{I},
-					{I}, {I}, {I},
+					{I},
+/* : getkey: Alt_Control_Space      0x139 */	{MARK},
+					{I},
 /* #define K_F1                   0x13b */	{F1},
 /* #define K_F2                   0x13c */	{F2},
 /* #define K_F3                   0x13d */	{F3},
@@ -295,11 +154,11 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Home                 0x147 */	{HOMEkey},
 /* #define K_Up                   0x148 */	{MUP},
 /* #define K_PageUp               0x149 */	{PU},
-/* #define K_Alt_KPMinus          0x14a */	{I},
+/* : #define K_Alt_KPMinus          0x14a */	{screenlesslines, alt_mask},
 /* #define K_Left                 0x14b */	{MLF},
 /* #define K_Center               0x14c */	{HOP},
 /* #define K_Right                0x14d */	{MRT},
-/* #define K_Alt_KPPlus           0x14e */	{I},
+/* : #define K_Alt_KPPlus           0x14e */	{screenmorelines, alt_mask},
 /* #define K_End                  0x14f */	{ENDkey},
 /* #define K_Down                 0x150 */	{MDN},
 /* #define K_PageDown             0x151 */	{PD},
@@ -341,16 +200,16 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Control_End          0x175 */	{ENDkey, ctrl_mask},
 /* #define K_Control_PageDown     0x176 */	{PD, ctrl_mask},
 /* #define K_Control_Home         0x177 */	{HOMEkey, ctrl_mask},
-/* #define K_Alt_1                0x178 */	{I},
-/* #define K_Alt_2                0x179 */	{I},
-/* #define K_Alt_3                0x17a */	{I},
-/* #define K_Alt_4                0x17b */	{I},
-/* #define K_Alt_5                0x17c */	{I},
-/* #define K_Alt_6                0x17d */	{I},
-/* #define K_Alt_7                0x17e */	{I},
-/* #define K_Alt_8                0x17f */	{I},
-/* #define K_Alt_9                0x180 */	{I},
-/* #define K_Alt_0                0x181 */	{I},
+/* : #define K_Alt_1                0x178 */	{key_1, alt_mask},
+/* : #define K_Alt_2                0x179 */	{key_2, alt_mask},
+/* : #define K_Alt_3                0x17a */	{key_3, alt_mask},
+/* : #define K_Alt_4                0x17b */	{key_4, alt_mask},
+/* : #define K_Alt_5                0x17c */	{key_5, alt_mask},
+/* : #define K_Alt_6                0x17d */	{key_6, alt_mask},
+/* : #define K_Alt_7                0x17e */	{key_7, alt_mask},
+/* : #define K_Alt_8                0x17f */	{key_8, alt_mask},
+/* : #define K_Alt_9                0x180 */	{key_9, alt_mask},
+/* : #define K_Alt_0                0x181 */	{key_0, alt_mask},
 /* #define K_Alt_Dash             0x182 */	{I},
 /* #define K_Alt_Equals           0x183 */	{I},
 /* #define K_Control_PageUp       0x184 */	{PU, ctrl_mask},
@@ -363,15 +222,15 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Alt_F11              0x18b */	{F11, alt_mask},
 /* #define K_Alt_F12              0x18c */	{F12, alt_mask},
 /* #define K_Control_Up           0x18d */	{MUP, ctrl_mask},
-/* #define K_Control_KPDash       0x18e */	{I},
+/* #define K_Control_KPDash       0x18e */	{screensmaller, ctrl_mask},
 /* #define K_Control_Center       0x18f */	{HOP, ctrl_mask},
-/* #define K_Control_KPPlus       0x190 */	{I},
+/* #define K_Control_KPPlus       0x190 */	{screenbigger, ctrl_mask},
 /* #define K_Control_Down         0x191 */	{MDN, ctrl_mask},
 /* #define K_Control_Insert       0x192 */	{INSkey, ctrl_mask},
 /* #define K_Control_Delete       0x193 */	{DELkey, ctrl_mask},
 					{I},
-/* #define K_Control_KPSlash      0x195 */	{I},
-/* #define K_Control_KPStar       0x196 */	{I},
+/* : #define K_Control_KPSlash      0x195 */	{FIND},
+/* : #define K_Control_KPStar       0x196 */	{GOHOP},
 /* #define K_Alt_EHome            0x197 */	{smallHOMEkey, alt_mask},
 /* #define K_Alt_EUp              0x198 */	{MUP, alt_mask},
 /* #define K_Alt_EPageUp          0x199 */	{PU, alt_mask},
@@ -385,8 +244,8 @@ struct pc_fkeyentry pc_xkey_map [512] =
 /* #define K_Alt_EPageDown        0x1a1 */	{PD, alt_mask},
 /* #define K_Alt_EInsert          0x1a2 */	{smallINSkey, alt_mask},
 /* #define K_Alt_EDelete          0x1a3 */	{smallDELkey, alt_mask},
-/* #define K_Alt_KPSlash          0x1a4 */	{I},
-/* #define K_Alt_Tab              0x1a5 */	{I},
+/* #define K_Alt_KPSlash          0x1a4 */	{FIND, alt_mask},
+/* #define K_Alt_Tab              0x1a5 */	{GOHOP},
 /* #define K_Alt_Enter            0x1a6 */	{SNL, alt_mask},
 					{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
@@ -439,11 +298,11 @@ struct pc_fkeyentry pc_xkey_map [512] =
 	{I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I}, {I},
 };
 
+
 #ifndef msdos
 
 /* mapping table: function key escape sequences -> mined functions */
-struct fkeyentry fkeymap [] =
-{
+struct fkeyentry fkeymap [] = {
 	/* Alt-Letter sequences for manual menu activation */
 /*
 	{"\033f",	FILEMENU},
@@ -453,6 +312,8 @@ struct fkeyentry fkeymap [] =
 	{"\033o",	OPTIONSMENU},
 */
 
+	{"\033]",	OSC},
+
 	/* Alt-Letter sequences for keyboard switching functions 
 	   to be recognised on prompt line */
 	{"\033k",	toggleKEYMAP},
@@ -645,10 +506,10 @@ struct fkeyentry fkeymap [] =
 
 	/* application keypad */
 	{"\033O ",	HOP},	/* Space, also old hanterm keypad 5 ?	*/
-	{"\033OI",	STAB},	/* shift-Tab */
+	{"\033OI",	HOP},	/* shift-Tab */
 	{"\033OM",	SNL},	/* Enter (@8) */
 	{"\033Oo",	FIND},	/* PF1, keypad /	*/
-	{"\033Oj",	I},	/* PF2, keypad *	*/
+	{"\033Oj",	GOHOP},	/* PF2, keypad *	*/
 	{"\033Om",	SU},	/* PF3, keypad -	*/
 	{"\033Ok",	SD},	/* PF4, keypad +	*/
 
@@ -665,7 +526,7 @@ struct fkeyentry fkeymap [] =
 	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
 	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
 
-	/* MinTTY generic patterns for modified application mode keys */
+	/* mintty generic patterns for modified application mode keys */
 	{"\033[1p",	INSkey},	/* Ins (kI), keypad 0	*/
 	{"\033[1q",	ENDkey},	/* End (K4), keypad 1	*/
 	{"\033[1r",	MDN},	/* dn (kd), keypad 2	*/
@@ -680,11 +541,11 @@ struct fkeyentry fkeymap [] =
 	{"\033[1l",	DELkey},	/* Del, keypad ,	*/
 	{"\033[1M",	SNL},	/* Enter (@8) */
 	{"\033[1o",	FIND},	/* keypad /	*/
-	{"\033[1j",	I},	/* keypad *	*/
-	{"\033[1m",	SU},	/* keypad -	*/
-	{"\033[1k",	SD},	/* keypad +	*/
+	{"\033[1j",	GOHOP},	/* keypad *	*/
+	{"\033[1m",	KP_minus},	/* keypad -	*/
+	{"\033[1k",	KP_plus},	/* keypad +	*/
 
-	/* MinTTY generic patterns for modified TAB */
+	/* mintty generic patterns for modified TAB */
 	{"\033[1Z",	STAB},	/* TAB	*/
 
 	/* Sun */
@@ -777,12 +638,11 @@ struct fkeyentry fkeymap [] =
 	{"\033OX",	F9},	/* F9, conflict with VT102 appl. mode =	*/
 	{"\033OY",	F10},	/* F10	*/
 	{"\033OZ",	F11},	/* F11	*/
-	{"\033O[",	F12},	/* F12	*/
 
 	/* rxvt */
 	{"\033[7~",	smallHOMEkey},	/* 	*/
 	{"\033[8~",	smallENDkey},	/* 	*/
-	{"\033[Z",	MPW},	/* shift-TAB, also MinTTY */
+	{"\033[Z",	HOP},	/* shift-TAB, also mintty */
 
 	/* SCO ansi */
 	{"\033[L",	INSkey},	/* (termcap: insert character) */
@@ -794,10 +654,13 @@ struct fkeyentry fkeymap [] =
 	{"\033[O",	FOCUSout},	/* window loses focus */
 	{"\033[I",	FOCUSin},	/* window gains focus */
 
-	/* MinTTY ambiguous width change report */
+	/* mintty ambiguous width change report */
 	{"\033[1W",	AMBIGnarrow},	/* ambig. width chars are narrow */
 	{"\033[2W",	AMBIGwide},	/* ambig. width chars are wide */
 
+	/* mintty application escape key mode (enabled with ^[[?7727h) */
+	{"\033O[",	ESCAPE},	/* ESC */
+
 	{NIL_PTR}
 };
 
@@ -829,9 +692,9 @@ struct fkeyentry fkeymap_vt52 [] = {
 	{"\033?M",	SNL},	/* keypad ENTER */
 
 	/* xterm/VT52 application mode */
-	{"\033?j",	I},	/* keypad * */
+	{"\033?j",	GOHOP},	/* keypad * */
 	{"\033?k",	SD},	/* keypad + */
-	{"\033?o",	I},	/* keypad / */
+	{"\033?o",	FIND},	/* keypad / */
 	{"\033?X",	HOP},	/* keypad = */
 	{"\033?I",	STAB},	/* Tab */
 	{"\033? ",	SSPACE},	/* Space ? */
@@ -948,6 +811,19 @@ struct fkeyentry fkeymap_rxvt [] = {
 	{"\033[23$",	F11, shift_mask},
 	{"\033[24$",	F12, shift_mask},
 
+	/* rxvt-unicode application keypad */
+	{"\033OU",	HOMEkey},	/* Pos1	*/
+	{"\033OV",	MLF},	/* lf	*/
+	{"\033OW",	MUP},	/* up	*/
+	{"\033OX",	MRT},	/* rt	*/
+	{"\033OY",	MDN},	/* dn	*/
+	{"\033OZ",	PU},	/* PgUp	*/
+	{"\033O]",	HOP},	/* 5	*/
+	{"\033O\\",	ENDkey},	/* End	*/
+	{"\033O[",	PD},	/* PgDn	*/
+	{"\033O^",	INSkey},	/* Ins	*/
+	{"\033O_",	DELkey},	/* Del	*/
+
 	{NIL_PTR}
 };
 
@@ -1004,6 +880,10 @@ struct fkeyentry fkeymap_vt100 [] = {
 	{"\033On",	DELkey},	/* Del (kD), keypad .	*/
 	{"\033Ol",	DELkey},	/* Del, keypad ,	*/
 
+	/* BSD console */
+	{"\033[7~",	HOMEkey},	/* 	*/
+	{"\033[8~",	ENDkey},	/* 	*/
+
 	{NIL_PTR}
 };
 
