@@ -3,9 +3,20 @@
 |*		keyboard mapping tables					*|
 \*======================================================================*/
 
+#ifdef vms
+#define NOSTRINGTABLES
+#define noCJKkeymaps
+#endif
+
+
 #include "mined.h"
 
 
+#ifdef msdos
+#define noCJKkeymaps
+#endif
+
+
 /*======================================================================*\
 |*		Keyboard mapping tables					*|
 \*======================================================================*/
@@ -15,7 +26,11 @@
 #ifdef keymapping_stringtables
 #include "_keymaps.h"
 #else
+# ifdef noCJKkeymaps
+#include "keymapsd.h"
+# else
 #include "keymapsi.h"
+# endif
 #endif
 
 #endif
@@ -25,10 +40,13 @@
 |*		Table of tables						*|
 \*======================================================================*/
 
-struct keymapping keymappingtable [] =
-{
+struct keymapping keymappingtable [] = {
 #ifdef use_keymap_tables
+# ifdef noCJKkeymaps
+#include "keymapsd.t"
+# else
 #include "keymapsk.t"
+# endif
 #else
 # ifdef __TURBOC__
 	{0, "--"},
