@@ -14,7 +14,7 @@
 /* MSDOS basic configuration */
 #ifdef __MSDOS__
 
-#undef unix	/* needed for djgcc? */
+#undef unix	/* needed for djgcc! */
 #define msdos
 #define pc
 #define pc_term
@@ -30,14 +30,18 @@
 #endif
 
 
+/* common includes */
+#include <stdio.h>
+#include <string.h>
+#include <unistd.h>
+
+
 /* MSDOS includes and declarations */
 #ifdef msdos
 extern void delay ();
 extern void exit ();
 
-#include <string.h>
 #include <io.h>		/* this is not "io.h" */
-#  include <stdio.h>
 #  undef putchar
 #  include <fcntl.h>
 
@@ -50,6 +54,7 @@ extern void sleep ();
 # endif
 
 # ifdef __TURBOC__
+#define time_t	unsigned long
 /* #define conio with combined positioning/output method...? */
 #define conio
 # else
@@ -93,19 +98,13 @@ extern void sleep ();
 /* Unix includes and defines */
 #ifdef unix
 
-/*#include <stdlib.h> */
-#  include <stdio.h>
-#  include <string.h>
-/*#include <strings.h> ? */
 #  include <fcntl.h>
-#  include <unistd.h>
 #  undef putchar
 
 #  ifdef sysV
    extern char * getcwd ();
 #  else
    extern char * getwd ();
-#  define getcwd(dirbuf, buflen)	getwd (dirbuf)
 #  endif
 
 extern void exit ();
@@ -170,8 +169,14 @@ extern unsigned int usleep ();
 
 /* CJK tables mode */
 #ifndef msdos
-#define use_cjk_tables
+# ifndef vms
+# define use_cjk_tables
+# endif
 #define use_keymap_tables
+#else
+# ifndef __TURBOC__
+# define use_keymap_tables
+# endif
 #endif
 
 
@@ -184,23 +189,15 @@ extern unsigned int usleep ();
 
 
 /***************************************************************************\
-	declaration of types and constants
+	declaration of types, basic macros, and constants
 \***************************************************************************/
 
 #define arrlen(arr)	(sizeof (arr) / sizeof (* arr))
 
 typedef unsigned char character;
-/*typedef unsigned int character;*/
+
 
 /* Screen size and display definitions. Coordinates start at 0, 0 */
-#ifdef pcxx
-#  define maxYMAX		64
-#  define maxXMAX		132
-#else
-#  define maxYMAX		126	/* 73 */
-#  define maxXMAX		279	/* 163 */
-#endif
-
 extern int YMAX;
 extern int XMAX;
 extern short MENU;
@@ -208,7 +205,21 @@ extern short MENU;
 #define XBREAK		(XMAX - scrollbar_width)	/* Shift line display at this column */
 #define XBREAK_bottom	(XMAX - 1)
 #define SHIFT_SIZE	(tab8 (XMAX / 4 + 1))	/* Number of chars to shift */
-#define maxLINE_LEN	(maxXMAX + 1)	/* max screen line length */
+
+/* Adjust buffer sizes, align with typical max screen width
+	screen size 1600x1200
+		font size 5x7:	172 x 319
+		font size 4x6:	200 x 399
+	screen size 2560x1600 (WQXGA+)
+		font size 4x6:	267 x 639
+   Restrain to reasonable size assumption (esp. @ reasonable font size), 
+   so prompt line buffer stays @ reasonable size < MAX_CHARS,
+   so let maxXMAX < MAX_CHARS / 3
+ */
+#define maxXMAX		319
+#define maxPROMPTlen	(maxXMAX + 1)	/* max screen line length, also used for prompt buffer */
+#define maxFILENAMElen	maxPROMPTlen
+#define maxCMDlen	2 * maxPROMPTlen + 9
 
 #define MAX_CHARS	1024		/* max chars on one line of text */
 		/* LINE_START must be rounded up to the lowest SHIFT_SIZE */
@@ -218,6 +229,7 @@ extern short MENU;
 					x-coordinate for line; MAX_CHARS + 1
 					is not sufficient due to TABs */
 
+
 /* Return values of functions */
 #define ERRORS		-1
 #define NO_LINE		(ERRORS - 1)	/* Must be < 0 */
@@ -232,11 +244,9 @@ extern short MENU;
 
 #define REPORT_CHANGED_LINES	1	/* Report change of lines on # lines */
 
-#ifndef pc
 #ifndef O_BINARY
 #define O_BINARY 0
 #endif
-#endif
 
 /*
  * mouse button selection
@@ -280,7 +290,7 @@ typedef enum {
 
 /* smart quotes state */
   UNSURE,
-  OPEN,
+  OPEN
 } FLAG;
 
 /*
@@ -304,11 +314,14 @@ struct Line {
   struct Line * next;
   struct Line * prev;
   char * text;
-  int highlighting_begin, highlighting_end;
-  unsigned short shift_count;
+  char * sel_begin;	/* beginning pos of selection, or NIL if none */
+  char * sel_end;	/* end pos of selection, or NIL if beyond line */
+  unsigned short shift_count;	/* horizontal display shifting */
   lineend_type return_type;
-  char syntax_marker;	/* bitmask composed of syntax_ flags below */
+  char syntax_mask;	/* bitmask composed of display flags below */
+  FLAG dirty;		/* flag that display attributes have changed */
 };
+/* display flags for struct Line */
 #define syntax_none	0
 #define syntax_HTML	1
 #define syntax_JSP	2
@@ -321,14 +334,6 @@ typedef struct Line LINE;
  */
 typedef void (* voidfunc) ();
 typedef int (* intfunc) ();
-#ifdef DEFPROTO_functions
-/* -> weird syntax error... */
-typedef void (* signalfunc) (int);
-typedef int (* flagfunc) (menuitemtype *, int);
-#else
-typedef void (* signalfunc) ();
-typedef int (* flagfunc) ();
-#endif
 
 /* NULL definitions */
 #define NIL_PTR		((char *) 0)
@@ -349,7 +354,7 @@ extern int access ();
 /*extern int open ();*/
 extern int close ();
 /*extern int creat ();*/
-extern int chdir ();
+/*extern int chdir ();*/
 extern int system ();
 extern int isatty ();
 #ifndef strcmp
@@ -361,6 +366,8 @@ extern int strcmp ();
 	variable declarations
 \***************************************************************************/
 
+extern char * debug_mined;	/* Debug indicator and optional flags */
+
 extern int total_lines;		/* Number of lines in file */
 extern long total_chars;	/* Number of characters in file */
 extern LINE * header;		/* Head of line list */
@@ -374,25 +381,25 @@ extern int line_number;		/* current line # determined by file_status */
 extern int lines_per_page;	/* assumption for file_status */
 
 extern int x, y;		/* x, y coordinates on screen */
+
+#define viewonly	(viewonly_mode || viewonly_locked || viewonly_err)
+extern FLAG viewonly_mode;	/* Set when view only mode is selected */
+extern FLAG viewonly_locked;	/* Enforced view only status */
+extern FLAG viewonly_err;	/* Error view only status */
 extern FLAG modified;		/* Set when file is modified */
-extern FLAG viewonly;		/* Set when view only mode is selected */
 extern FLAG restricted;		/* Set when restricted mode is selected */
-extern FLAG quit;		/* Set when quit character is typed */
-extern FLAG intr_char;		/* Set when intr character is typed */
+extern FLAG quit;		/* Set on SIGQUIT or quit character typed */
 extern FLAG winchg;		/* Set when the window size changes */
 extern FLAG interrupted;	/* Set when a signal interrupts */
 extern FLAG waitingforinput;	/* Set while waiting for the next command key */
 extern FLAG isscreenmode;	/* Set when screen mode is on */
 extern char text_buffer [MAX_CHARS];	/* Buffer for modifying text */
-extern int input_fd;		/* File descriptors for terminal dialog */
-extern int output_fd;
 
 extern char file_name [];	/* name of file being edited */
 extern FLAG writable;		/* Set if file cannot be written */
 extern FLAG file_is_dir;	/* Tried to open a directory as file? */
 extern FLAG reading_pipe;	/* Set if file should be read from stdin */
 extern FLAG writing_pipe;	/* Set if file should be written to stdout */
-extern lineend_type default_lineend;
 
 extern PROT xprot;		/* actual prot. mask representing +x option */
 extern PROT fprot0;		/* default prot. mode for new files */
@@ -400,8 +407,7 @@ extern PROT fprot1;		/* prot. mode for new file being edited */
 extern PROT bufprot;		/* prot. mode for paste buffer file */
 
 #ifdef unix
-extern char * window_string_code;
-extern char window_string [];	/* to set text in window headline and icon */
+extern char * title_string_pattern;
 #endif
 
 extern char yank_file [];	/* Temp. file for buffer */
@@ -419,11 +425,21 @@ extern int buffer_open_flag;	/* counter flag for the collective buffer */
 extern int JUSlevel;		/* keep justified while typing? */
 extern int JUSmode;		/* 1: paragraphs end at empty line */
 extern FLAG autoindent;		/* auto indent on input of Enter? */
-extern FLAG dim_HTML;		/* display HTML dimmed ? */
+extern FLAG mark_HTML;		/* display HTML marked/dimmed ? */
 extern FLAG insert_mode;	/* insert or overwrite */
 extern FLAG append_flag;	/* Set when buffer should be appended to */
 extern FLAG pastebuf_utf8;	/* Paste buffer always treated as UTF-8? */
 extern FLAG rectangular_paste_flag;	/* Copy/Paste rectangular area? */
+extern FLAG visselect_key;	/* Visual selection on unmodified cursor key? */
+extern FLAG visselect_anymouse;		/* Visual selection on any mouse move? */
+extern FLAG visselect_keeponsearch;	/* Keep visual selection on search ? */
+extern FLAG visselect_setonfind;	/* Select search result? */
+extern FLAG visselect_autocopy;		/* Auto-copy selection? */
+extern FLAG visselect_autodelete;	/* Delete selection on insert? */
+
+extern char backup_mode;	/* none, simple, numbered (e/v), automatic */
+extern char * backup_directory;	/* directory name for backup files */
+extern char * recover_directory;	/* directory name for recovery files */
 extern character erase_char;	/* effective (configured) char for erase left */
 extern char emulation;		/* 'w' for WordStar, 'e' for emacs */
 extern FLAG emacs_buffer;	/* enable emacs buffer fct for ^K/^T */
@@ -464,13 +480,13 @@ extern FLAG poormansbidi;	/* poor man's bidirectional support ? */
 
 extern FLAG combining_mode;	/* UTF-8 combining character support ? */
 extern FLAG separate_isolated_combinings;	/* separated display of comb. at line beg./after TAB ? */
+extern FLAG apply_joining;	/* apply LAM/ALEF joining ? */
 
 extern FLAG disp_scrollbar;	/* shall scrollbar be displayed ? */
 extern FLAG fine_scrollbar;	/* fine-grained UTF-8 scrollbar ? */
 extern FLAG old_scrollbar_clickmode;	/* old left/right click semantics ? */
 extern int scrollbar_width;
 extern FLAG update_scrollbar_lazy;	/* partial scrollbar refresh as needed ? */
-extern FLAG no_window_title;	/* suppress filename display in window title? */
 
 extern FLAG loading;		/* Loading a file? Init True for error handling */
 extern FLAG only_detect_text_encoding;
@@ -514,6 +530,9 @@ extern char * UTF_RETfini_marker;	/* Char to fill last position of line with */
 extern char * UTF_SHIFT_marker;		/* Char indicating that line continues */
 extern char * UTF_SHIFT_BEG_marker;	/* Char indicating that line continues left */
 extern char * UTF_MENU_marker;	/* Char to mark selected item */
+extern char * submenu_marker;	/* Char to mark submenu entry */
+extern char * menu_cont_marker;	/* Char to mark menu border continuation */
+extern char * menu_cont_fatmarker; /* Char to mark menu border continuation */
 
 extern FLAG stat_visible;	/* Set if status line is visible */
 extern FLAG top_line_scrolled;	/* Was menu line scrolled away? */
@@ -546,6 +565,7 @@ extern void move_address_w_o_RD _((char *, int y));
 extern int find_x _((LINE * line, char * address));
 extern int get_cur_pos _((void));
 extern int get_cur_col _((void));
+extern int get_text_col _((LINE *, char *));
 /* from edit.c */
 extern int find_y _((LINE * line));
 extern int find_y_w_o_RD _((LINE * line));
@@ -557,30 +577,33 @@ extern void set_modified _((void));
 extern void file_status _((char * message, long bytecount, long charcount, 
 			char * filename, int lines, FLAG textstat, 
 			FLAG writefl, FLAG changed, FLAG viewing));
-extern void calc_line_no _((void));
+
+/* from edit.c */
+extern char syntax_state _((char prev_mask, char old_mask, char * spoi, char * sbeg));
+extern void update_text_state _((LINE *));
 
 /* from pastebuf.c */
 extern void delete_yank_files _((void));
-extern int scratchfile _((FLAG read_write, FLAG append, char * buf_name, char * which, FLAG * buf_status));
 extern int yankfile _((FLAG read_write, FLAG append));
-
-/* from edit.c */
-extern char syntax_state _((char, char *));
-extern void update_text_state _((LINE *));
-
-/* paste buffer functions */
-/* from pastebuf.c */
+extern void yank_HTML _((FLAG remove));
+extern void paste_HTML _((void));
+extern FLAG checkmark _((LINE *, char *));
 extern void delete_text_buf _((LINE * start_line, char * start_textp, LINE * end_line, char * end_textp));
 
+/* selection highlighting */
+/* from pastebuf.c */
+extern void start_highlight_selection _((void));
+extern void continue_highlight_selection _((void));
+extern void clear_highlight_selection _((void));
+extern FLAG has_active_selection _((void));
+extern void update_selection_marks _((int pos_x));
+
 /* editing functions */
 /* from edit.c */
 extern int delete_text _((LINE * startl, char * start, LINE * endl, char * end));
+extern int delete_text_only _((LINE * startl, char * start, LINE * endl, char * end));
 extern int insert_text _((LINE *, char * location, char * string));
-extern LINE * line_insert _((LINE *, char *, int, lineend_type));
-
-/* from pastebuf.c */
-extern void yank_HTML _((FLAG remove));
-extern void paste_HTML _((void));
+extern LINE * line_insert_after _((LINE *, char *, int, lineend_type));
 
 /* character input handling */
 struct prefixspec {
@@ -595,7 +618,7 @@ struct prefixspec {
 
 /* from compose.c */
 extern char * mnemos _((int ucs));
-extern unsigned long compose _((unsigned long, unsigned long));
+extern unsigned long compose_chars _((unsigned long, unsigned long));
 extern unsigned long compose_mnemonic _((char *));
 extern struct prefixspec * lookup_prefix _((voidfunc prefunc, unsigned int shift));
 extern struct prefixspec * lookup_prefix_char _((unsigned long c));
@@ -613,6 +636,7 @@ extern void key_6 _((void));
 extern void key_7 _((void));
 extern void key_8 _((void));
 extern void key_9 _((void));
+extern unsigned long CTRLGET _((FLAG check_prefix));
 
 
 /* status line handling */
@@ -620,7 +644,10 @@ extern void key_9 _((void));
 extern void rd_bottom_line _((void));
 extern void redraw_prompt _((void));
 extern void clear_lastline _((void));
-extern int get_filename _((char * message, char * filename));
+/* file chooser */
+extern int get_filename _((char * message, char * filename, FLAG directory));
+extern FLAG sort_by_extension;
+extern FLAG sort_dirs_first;
 
 
 /*
@@ -650,23 +677,51 @@ extern void search_corresponding _((char *, FLAG, char *));
 extern void viewonlyerr _((void));
 extern void restrictederr _((void));
 
+extern void emul_emacs _((void));
+extern void emul_WordStar _((void));
+extern void emul_pico _((void));
+extern void emul_Windows _((void));
+extern void emul_mined _((void));
+
+extern void convlineend_cur_LF _((void));
+extern void convlineend_cur_CRLF _((void));
+extern void convlineend_all_LF _((void));
+extern void convlineend_all_CRLF _((void));
+
 
 /* special stuff */
 /* from minedaux.c */
+extern void debuglog _((char * tag, char * l1, char * l2));
 extern char * envvar _((char * name));
-extern void catch_interrupt _((int));
+extern char * envstr _((char * name));
+extern char * getbasename _((char * s));
+extern int is_absolute_path _((char * filename));
+extern char * getusername _((void));
+extern char * get_cwd _((char * dirbuf));
+extern char * gethomedir _((void));
+extern void sethomedir _((char *));
+extern int systemcall _((char * msg, char * cmd, int delay));
+#ifndef msdos
+extern int progcallpp _((char * msg, int delay, char * binprepath [], 
+	char * dir, char * prog, char * p1, char * p2, char * p3, char * p4));
+#endif
+#ifdef msdos
+extern int is_Windows _((void));
+#endif
 extern void panic _((char *, char *));
 extern void panicio _((char * message, char * err));
-extern void bad_write _((int fd));
 extern char * scan_int _((char *, int *));
 extern char * dec_out _((long));
 extern char * serror _((void));
 extern char * serrorof _((int));
 extern int geterrno _((void));
-extern void delete_file _((char *));
+extern int delete_file _((char *));
+extern FLAG copyfile _((char * from_file, char * to_file));
 extern char * unnull _((char *));
 extern char * num_out _((long number, long radix));
 extern void copy_string _((char * to, char * from));
+/* from legacy.c */
+extern int portable_getpgrp _((int mypid));
 
 /* memory allocation */
 /* from minedaux.c */
@@ -682,7 +737,6 @@ extern voidfunc command _((unsigned long));
 
 /* prompt line */
 /* from prompt.c */
-extern character promptyn _((void));
 extern FLAG char_on_status_line; /* is output active on status line ? */
 extern FLAG input_active;
 extern int lpos;
@@ -702,6 +756,7 @@ extern FLAG iscombined_unichar _((unsigned long ucs, char * charpos, char * line
 extern unsigned long case_convert _((unsigned long unichar, int dir));
 /* from charcode.c */
 extern int iscontrol _((unsigned long));
+extern int iswhitespace _((unsigned long));
 extern character controlchar _((character));
 
 /* scrollbar handling */
@@ -716,11 +771,13 @@ extern void clear_filebuf _((void));
 extern int flush_filebuf _((int fd));
 /* from mined1.c */
 extern void RDwin _((void));
+extern void RDwinchg _((void));
 extern void RD_window_title _((void));
 extern void clear_window_title _((void));
 extern void RD_y _((int y_pos));
 /* from output.c */
 extern void display _((int y, LINE *, int count, int new_pos));
+extern void display_flush _((void));
 /* from output.c */
 extern void putmark _((char mark, char * utfmark));
 extern FLAG marker_defined _((character marker, char * utfmarker));
@@ -737,6 +794,10 @@ extern int put_cjkchar _((unsigned long));
 /* character handling */
 /* from minedaux.c */
 extern int length_of _((char *));
+extern char * dupstr _((char * s));
+extern char * nextutf8 _((char *s));
+extern int stringcompare_ _((char * s1, char * s2));
+extern int stringprefix_ _((char * s1, char * s2));
 /* from mined1.c */
 extern int char_count _((char *));
 extern int col_count _((char *));
@@ -761,6 +822,7 @@ extern int iscombined _((unsigned long, char *, char *));
 #define multichar(c)	((character) c >= 0x80 && (cjk_text == False || text_encoding_tag != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
 extern int iscombining _((unsigned long ucs));
 extern int iswide _((unsigned long ucs));
+extern int isglyph_uni _((unsigned long));
 
 /* from isrtol.c */
 extern int is_right_to_left _((unsigned long ucs));
@@ -773,6 +835,7 @@ extern char * encode_char _((unsigned long));
 extern FLAG valid_cjk _((unsigned long cjkchar, /* opt */ character * cjkbytes));
 extern FLAG valid_cjkchar _((FLAG term, unsigned long cjkchar, /* opt */ character * cjkbytes));
 extern unsigned long isolated_alef _((unsigned long));
+extern unsigned long ligature_lam_alef _((unsigned long));
 extern FLAG no_char _((unsigned long c));
 extern FLAG no_unichar _((unsigned long u));
 
@@ -815,28 +878,18 @@ extern FLAG disp_Han_Hangul;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Korean;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Vietnamese;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_HanyuPinlu;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_HanyuPinyin;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_XHCHanyuPinyin;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_Tang;	/* display this Han pronunciation ? */
 extern FLAG disp_Han_description;	/* display Han description ? */
 extern FLAG disp_Han_full;	/* display full popup Han description ? */
 
-/* Han character descriptions table */
-struct hanentry {
+struct raw_hanentry {
 	unsigned long unicode;
-	char * Mandarin;
-	char * Cantonese;
-	char * Japanese;
-	char * Sino_Japanese;
-	char * Hangul;
-	char * Korean;
-	char * Vietnamese;
-	char * HanyuPinlu;
-	char * XHCHanyuPinyin;
-	char * Tang;
-	char * Definition;
+	char * text;
 };
 
-extern struct hanentry hantable [];
+extern struct raw_hanentry hantable [];
 extern unsigned int hantable_len;
 
 
@@ -875,17 +928,21 @@ extern FLAG enforce_keymap;	/* enable keyboard mapping even on non-suitable term
 #define SPACE_SELECT	's'
 
 /* menu handling */
+typedef void (* _menufunc) ();
+typedef int (* _flagfunc) ();
 typedef struct {
 	char * itemname;
-	voidfunc itemfu;
+	_menufunc itemfu;
 	char * hopitemname;
-	flagfunc itemon;
+	_flagfunc itemon;
 	char * extratag;
 	} menuitemtype;
-extern void separator _((void));
+typedef void (* menufunc) _((menuitemtype *, int));
+typedef int (* flagfunc) _((menuitemtype *, int));
+
 extern FLAG flags_changed;	/* Should flag menu area be redrawn? */
 /* from mousemen.c */
-extern void fill_menuitem _((menuitemtype * item, char * s));
+extern void fill_menuitem _((menuitemtype * item, char * s1, char * s2));
 extern void openmenuat _((int col));
 extern int popup_menu _((menuitemtype *, int menulen, int column, int line, 
 			char * title, 
@@ -920,9 +977,10 @@ extern void invoke_key_function _((unsigned long));
 extern void QUICKMENU _((void)), FILEMENU _((void)), EDITMENU _((void));
 extern void SEARCHMENU _((void)), PARAMENU _((void)), OPTIONSMENU _((void));
 extern void handleFlagmenus _((void));
-extern void MOUSEescape _((void)), FOCUSout _((void)), FOCUSin _((void));
+extern void MOUSEfunction _((void)), FOCUSout _((void)), FOCUSin _((void));
 extern void AMBIGnarrow _((void)), AMBIGwide _((void));
 extern void ANSIseq _((void));
+extern void OSC _((void));
 extern void MUP _((void)), MDN _((void)), MLF _((void)), MRT _((void));
 extern void MNW _((void)), MPW _((void));
 extern void BSEN _((void)), ESEN _((void));
@@ -936,7 +994,9 @@ extern void DCC0 _((void)), DPC0 _((void));
 extern void CTRLINS _((void)), DLINE _((void)), TOGINS _((void)), ctrlQ _((void)), ctrlK _((void)), ctrlO _((void));
 extern void search_wrong_enc _((void));
 extern void JUS _((void)), JUSclever _((void)), JUSandreturn _((void));
-extern void QUED _((void)), WT _((void)), WTU _((void)), SAVEAS _((void)), SAVPOS _((void)), EDIT _((void)), NN _((void));
+extern void QUED _((void)), WT _((void)), WTU _((void)), SAVEAS _((void)), EDIT _((void)), NN _((void));
+extern void RECOVER _((void));
+extern void SAVPOS _((void)), GROOM_INFO _((void));
 extern void RD _((void)), RDwin _((void)), RDwin_nomenu _((void)), I _((void));
 extern void EXED _((void)), VIEW _((void));
 extern void HOP _((void)), CANCEL _((void)), ESCAPE _((void));
@@ -952,6 +1012,7 @@ extern void ADJLM _((void)), ADJFLM _((void)), ADJNLM _((void)), ADJRM _((void))
 extern void toggle_tabsize _((void)), toggle_tab_expansion _((void));
 extern void changeuni _((void)), changehex _((void)), changeoct _((void)), changedec _((void));
 extern void screensmaller _((void)), screenbigger _((void)), screenlesslines _((void)), screenmorelines _((void));
+extern void fontsmaller _((void)), fontbigger _((void)), fontdefault _((void));
 
 extern void SFW _((void)), SRV _((void)), RS _((void)), RSreverse _((void));
 extern void LR _((void)), GR _((void)), REPL _((void));
@@ -967,15 +1028,17 @@ extern void S0 _((character)), Scharacter _((unsigned long));
 extern FLAG insert_unichar _((unsigned long));
 extern void SNL _((void)), STAB _((void)), SSPACE _((void));
 extern void LIB _((void)), Underline _((void));
-extern void Sdoublequote _((void)), Ssinglequote _((void)), Sdash _((void));
+extern void Sdoublequote _((void)), Ssinglequote _((void)), Sdash _((void)), Sacute _((void)), Sgrave _((void));
 extern void SCURCHARFW _((void)), SCURCHARRV _((void));
 
 extern void goline _((int)), goproz _((int));
 
-extern void MARK _((void)), COPY _((void)), CUT _((void)), toggleMARK _((void));
-extern void Pushmark _((void)), Popmark _((void));
-extern void GOTO _((void)), GOMA _((void));
-extern void MARKn _((int)), GOMAn _((int)), Markn _((int));
+extern void MARK _((void)), COPY _((void)), CUT _((void));
+extern void setMARK _((FLAG set_only)), toggleMARK _((void));
+extern void toggle_rectangular_paste_mode _((void)), SELECTION _((void));
+extern void Pushmark _((void)), Popmark _((void)), Upmark _((void));
+extern void GOTO _((void)), GOHOP _((void)), GOMA _((void));
+extern void MARKn _((int)), GOMAn _((int)), mark_n _((int));
 extern void PASTE _((void)), PASTEEXT _((void)), PASTEstay _((void)), YANKRING _((void));
 extern void WB _((void)), INSFILE _((void));
 
@@ -988,17 +1051,21 @@ extern void SCORR _((FLAG pref_direction));
 extern void checkout _((void)), checkin _((void));
 extern void SPELL _((void));
 
-extern void FS _((void)), FSTAT _((void));
+extern void FS _((void)), FSTATUS _((void));
 extern void display_code _((void)), display_the_code _((void));
 extern void display_Han _((char * cpoi, FLAG force_utf8));
 extern void display_FHELP _((void));
 extern void ABOUT _((void));
 extern void HELP _((void)), HELP_topics _((void)), HELP_Fn _((void)), HELP_accents _((void)), toggle_FHELP _((void));
+
 extern void FHELP _((voidfunc));
 
+extern void check_file_modified _((void));
+
 /* function keys */
 extern void HOMEkey _((void)), ENDkey _((void)), smallHOMEkey _((void)), smallENDkey _((void));
 extern void INSkey _((void)), DELkey _((void)), smallINSkey _((void)), smallDELkey _((void));
+extern void KP_plus _((void)), KP_minus _((void));
 extern void DELchar _((void));
 extern void FIND _((void)), AGAIN _((void));
 extern void F1 _((void)), F2 _((void)), F3 _((void)), F4 _((void)), F5 _((void)), F6 _((void)), F7 _((void)), F8 _((void)), F9 _((void)), F10 _((void)), F11 _((void)), F12 _((void));
@@ -1008,14 +1075,17 @@ extern void F1 _((void)), F2 _((void)), F3 _((void)), F4 _((void)), F5 _((void))
  * String functions
  */
 #define streq(s1, s2)		(strcmp (s1, s2) == 0)
-#define strisprefix(s1, s2)	(strncmp (s2, s1, strlen (s1)) == 0)
-#define strcontains(s1, s2)	(strstr (s1, s2))
+#define strisprefix(p, s)	(strncmp (s, p, strlen (p)) == 0)
+#define strcontains(s, c)	(strstr (s, c))
+#define strappend(s, a, s_size)	strncat (s, a, s_size - strlen (s) - 1)
 
 /*
  * Functions handling status_line.
  */
 extern int bottom_line _((FLAG attrib, char *, char *, char * inbuf, FLAG statfl, char * term_input));
-extern void status_uni _((char *));
+extern void status_uni _((char * msg));
+extern character status2_prompt _((char * term, char * msg1, char * msg2));
+extern character prompt _((char * term));
 #define status_msg(str)		status_line (str, NIL_PTR)
 #define status_line(str1, str2)	\
 	(void) bottom_line (True, (str1), (str2), NIL_PTR, False, "")
@@ -1031,7 +1101,6 @@ extern void status_uni _((char *));
 	(void) bottom_line (False, NIL_PTR, NIL_PTR, NIL_PTR, False, "")
 #define get_string(str1, str2, fl, term_chars)	\
 	bottom_line (True, (str1), NIL_PTR, (str2), fl, term_chars)
-
 extern int get_string_uni _((char * str1, char * str2, FLAG fl, char * term_chars));
 
 
@@ -1043,12 +1112,8 @@ extern int get_string_uni _((char * str1, char * str2, FLAG fl, char * term_char
 
 /*
  * Build formatted string.
- * If this definition is left out, a routine will be defined in mined1.c.
  */
 #define build_string sprintf
-#ifndef build_string
-extern void build_string ();
-#endif
 
 
 /***************************************************************************\
