@@ -36,7 +36,8 @@ CFLAGS	= $(SYSFLAGS) $(CCFLAGS) $(OPT) $(DEBUG)
 
 LINKOPTS=$(LDOPTS)
 # for newer ncursesw?
-LDL=-ldl
+#LDL=-ldl	# if libdl library exists:
+LDL=` echo "/usr/lib ${LDFLAGS} " | sed -e "s, -L, ,g" -e "s,\([^ ][^ ]*\),ls -1 \1/libdl.*;,g" | sh | sed -e "s,.*/libdl\..*,-ldl," -e 1q `
 
 
 #############################################################################
