@@ -179,13 +179,9 @@ fine_grained_scrollbar (force)
   long last_line_number;
   int disp_start;
   int disp_end;
-  int i;
-  int scri;
   FLAG painted = False;
   int slices = 0;
   int oldslices = 0;
-  FLAG klopsinslot;
-  FLAG oldklopsinslot;
 
   screen_lines = last_y + 1;
   fields = screen_lines * unit;
@@ -222,6 +218,7 @@ fine_grained_scrollbar (force)
 #endif
 
   if (disp_scrollbar) {
+	int i;
 #ifdef debug_scrollbar
 	if (update_scrollbar_lazy) {
 		printf ("scrollbar (%d) %d - %d (prev. %d - %d)\n", force, disp_start, disp_end, prev_disp_start, prev_disp_end);
@@ -230,6 +227,8 @@ fine_grained_scrollbar (force)
 
 /* last_y / SCREENMAX */
 	for (i = 0; i < fields; i ++) {
+		FLAG klopsinslot;
+		FLAG oldklopsinslot;
 		if (i >= disp_start && i <= disp_end) {
 			slices ++;
 			klopsinslot = True;
@@ -243,7 +242,7 @@ fine_grained_scrollbar (force)
 			oldklopsinslot = False;
 		}
 		if (((i + 1) % unit) == 0) {
-		    scri = i / unit;
+		    int scri = i / unit;
 		    if (slices != oldslices 
 			|| klopsinslot != oldklopsinslot
 			|| force
@@ -332,11 +331,10 @@ cell_grained_scrollbar (force)
 /* last_y / YMAX */
   int disp_start = ((long) line_number - y) * last_y / total_lines;
   int disp_end = ((long) line_number - y + last_y + 1) * last_y / total_lines;
-  int i;
   FLAG painted = False;
-  FLAG isin_newklops, isin_oldklops;
 
   if (disp_scrollbar) {
+	int i;
 #ifdef debug_scrollbar
 	if (update_scrollbar_lazy) {
 		printf ("scrollbar (%d) %d - %d (prev. %d - %d)\n", force, disp_start, disp_end, prev_disp_start, prev_disp_end);
@@ -345,18 +343,20 @@ cell_grained_scrollbar (force)
 
 /* last_y / SCREENMAX */
 	for (i = 0; i <= last_y; i ++) {
-	    isin_newklops = i >= disp_start && i <= disp_end;
-	    isin_oldklops = i >= prev_disp_start && i <= prev_disp_end;
+	    FLAG isin_newklops = i >= disp_start && i <= disp_end;
+	    FLAG isin_oldklops = i >= prev_disp_start && i <= prev_disp_end;
 	    if (isin_newklops != isin_oldklops || force) {
 		painted = True;
 		set_cursor (XMAX, i);
 		if (isin_newklops) {
 #ifdef CJKterm_not_coloured
-			if (cjk_term) {
+			if (cjk_term)
+			{
 				reverse_on ();
 				putchar ('<');
 				reverse_off ();
-			} else
+			}
+			else
 #endif
 			{
 				disp_scrollbar_foreground ();
@@ -417,7 +417,7 @@ display_scrollbar (update)
 
 
 /*======================================================================*\
-|*			Screen function					*|
+|*			Screen functions				*|
 \*======================================================================*/
 
 void
@@ -430,6 +430,230 @@ clearscreen ()
 }
 
 
+/**
+   Manage nested/combined invocations of
+	combiningdisp_on/off
+	ctrldisp_on/off
+	unidisp_on/off
+	specialdisp_on/off
+	mark_on/off
+	disp_script ();
+		disp_colour ();
+		disp_normal ();
+	disp_syntax ();
+		dispHTML_jsp ();
+		dispHTML_comment ();
+		dispHTML_code ();
+		dispHTML_off ();
+	reverse_on/off (for selection highlighting)
+ */
+
+#define dont_debug_attr
+
+/* screen attributes are listed in order of precedence (increasing) */
+#define attr_html	0x001
+#define attr_comment	0x002
+#define attr_jsp	0x004
+
+#define attr_script	0x008
+
+#define attr_uni	0x010
+#define attr_special	0x020
+#define attr_mark	0x040
+#define attr_combining	0x080
+#define attr_control	0x100
+
+#define attr_selected	0x200
+
+#define attr_all	0xFFF
+
+static unsigned short attr = 0;
+static int attr_colour = 0;
+
+static
+void
+do_set_attr (unsigned short set)
+{
+#ifdef debug_attr
+  printf ("do_set_attr %03X\n", set);
+#endif
+  /* set attribute on screen unless preceded by an overriding one */
+  switch (set) {
+	case attr_combining:
+		combiningdisp_on ();
+		return;
+	case attr_uni:
+		unidisp_on ();
+		return;
+	case attr_mark:
+		mark_on ();
+		return;
+	case attr_special:
+		specialdisp_on ();
+		return;
+	case attr_script:
+		disp_colour (attr_colour);
+		return;
+	case attr_html:
+		dispHTML_code ();
+		return;
+	case attr_jsp:
+		dispHTML_jsp ();
+		return;
+	case attr_comment:
+		dispHTML_comment ();
+		return;
+	case attr_control:
+		if (attr & attr_selected) {
+			disp_normal ();
+		} else {
+			ctrldisp_on ();
+		}
+		return;
+	case attr_selected:
+		if (attr & attr_control) {
+			disp_normal ();
+		} else {
+			reverse_on ();
+		}
+		return;
+    }
+}
+
+static
+void
+refresh_attrs ()
+{
+  /* combine selected attributes and refresh them on screen */
+
+  if (attr & attr_uni) {
+	do_set_attr (attr_uni);
+  } else if (attr & attr_mark) {
+	do_set_attr (attr_mark);
+  } else if (attr & attr_special) {
+	do_set_attr (attr_special);
+  } else if (attr & attr_combining) {
+	do_set_attr (attr_combining);
+  } else if (attr & attr_control && ! (attr & attr_selected)) {
+	do_set_attr (attr_control);
+  } else {
+	if (attr & attr_jsp) {
+		do_set_attr (attr_jsp);
+	} else if (attr & attr_comment) {
+		do_set_attr (attr_comment);
+	} else if (attr & attr_html) {
+		do_set_attr (attr_html);
+	}
+	if (attr & attr_script) {
+		do_set_attr (attr_script);
+	}
+  }
+
+  if (attr & attr_selected && ! (attr & attr_control)) {
+	do_set_attr (attr_selected);
+  }
+}
+
+static
+void
+clear_attr (unsigned short clear)
+{
+#ifdef debug_attr
+  printf ("clear_attr %03X (%03X->%03X)\n", clear, attr, attr & ~ clear);
+#endif
+  attr &= ~ clear;
+
+  /* clear all attributes on screen */
+  disp_normal ();
+
+  refresh_attrs ();
+}
+
+static
+void
+set_attr (unsigned short set)
+{
+#ifdef debug_attr
+  printf ("set_attr %03X (%03X->%03X)\n", set, attr, attr | set);
+#endif
+  if (set > attr || set == attr_script || set == attr_selected) {
+	do_set_attr (set);
+	attr |= set;
+  } else if ((attr | set) != attr) {
+	attr |= set;
+	refresh_attrs ();
+  } else {
+	attr |= set;
+  }
+}
+
+static
+void
+clear_colour_attr ()
+{
+  clear_attr (attr_script);
+}
+
+static
+void
+set_colour_attr (text_colour)
+  int text_colour;
+{
+  attr_colour = text_colour;
+#ifdef debug_attr
+  printf ("set_colour_attr %d\n", attr_colour);
+#endif
+
+  set_attr (attr_script);
+}
+
+
+/*======================================================================*\
+|*			Character display substitution			*|
+\*======================================================================*/
+
+static FLAG glyphs_checked = False;
+static FLAG glyphs_available = False;
+static struct {
+	unsigned long unichar;
+	unsigned long subst;
+	FLAG checked;
+} glyphs_subst [] = {
+	{0x20AC, 'E', False},	/* € */
+	{0x2713, 'V', False},	/* ✓ */
+	{0x2717, 'X', False},	/* ✗ */
+	{0x21AF, 'Z', False},	/* ↯ */
+	{0x02BB, '\'', False}	/* glottal stop ʻokina */
+};
+
+static
+unsigned long
+disp_subst (unichar)
+  unsigned long unichar;
+{
+  if (! glyphs_checked) {
+	glyphs_available = isglyph_uni (0);
+	glyphs_checked = True;
+  }
+  if (glyphs_available) {
+	int i;
+	for (i = 0; i < arrlen (glyphs_subst); i ++) {
+		if (glyphs_subst [i].unichar == unichar) {
+			if (glyphs_subst [i].checked == False) {
+				if (isglyph_uni (unichar)) {
+					/* do not substitute */
+					glyphs_subst [i].subst = 0;
+				}
+				glyphs_subst [i].checked = True;
+			}
+			return glyphs_subst [i].subst;
+		}
+	}
+  }
+  return 0;
+}
+
+
 /*======================================================================*\
 |*			Marked character output				*|
 \*======================================================================*/
@@ -510,20 +734,21 @@ putnarrowutf (utfmarker)
  */
 static
 void
-indicate_uni (c, allow_wide)
+indicate_uni (c, allow_wide, attr)
   unsigned long c;
   FLAG allow_wide;
+  unsigned short attr;
 {
   if (char_on_status_line) {
 	reverse_off ();
   }
-  unidisp_on ();
+  set_attr (attr);
   if (allow_wide) {
 	put_unichar (c);
   } else {
 	putnarrowchar (c);
   }
-  unidisp_off ();
+  clear_attr (attr);
   if (char_on_status_line) {
 	reverse_on ();
   }
@@ -540,9 +765,9 @@ indicate_special (c)
   if (char_on_status_line) {
 	reverse_off ();
   }
-  unimarkdisp_on ();
+  set_attr (attr_special);
   putnarrowchar ((unsigned long) c);
-  unimarkdisp_off ();
+  clear_attr (attr_special);
   if (char_on_status_line) {
 	reverse_on ();
   }
@@ -558,7 +783,7 @@ indicate_special (c)
 
 #define indicate_illegal_UTF 	trace_indicate	indicate_marked
 #define indicate_narrow 	trace_indicate	indicate_marked
-#define indicate_marked(c) 	trace_indicate	indicate_uni(c, False)
+#define indicate_marked(c) 	trace_indicate	indicate_uni(c, False, attr_uni)
 
 
 /*
@@ -577,7 +802,7 @@ putmarkmode (marker, utfmarker, dim_me)
   FLAG dim_me;
 {
   if (dim_me && ! mark_mode_active) {
-	mark_on ();
+	set_attr (attr_mark);
 	mark_mode_active = True;
   }
 
@@ -624,7 +849,7 @@ void
 endmarkmode ()
 {
   if (mark_mode_active) {
-	mark_off ();
+	clear_attr (attr_mark);
 	mark_mode_active = False;
   }
   if (mark_alt_cset) {
@@ -651,27 +876,43 @@ putUmark (marker, utfmarker)
   character marker;
   char * utfmarker;
 {
-  unimarkdisp_on ();
+  set_attr (attr_special);
   putmarkmode (marker, utfmarker, False);
   endmarkmode ();
-  unimarkdisp_off ();
+  clear_attr (attr_special);
 }
 
 void
 put_menu_marker ()
 {
-/*
-	unidisp_on ();
-	If utf8_screen
-	Then	put_unichar (menu_marker);
-	Else	putchar ('*');
-	Fi
-	unidisp_off ();
-*/
-  unidisp_on ();
+  set_attr (attr_uni);
   putmarkmode (MENU_marker, UTF_MENU_marker, False);
   endmarkmode ();
-  unidisp_off ();
+  clear_attr (attr_uni);
+}
+
+void
+put_submenu_marker (with_attr)
+  FLAG with_attr;
+{
+  if (with_attr) {
+	set_attr (attr_uni);
+  }
+  if (utf8_screen) {
+	put_unichar (utf8value (submenu_marker));
+  } else {
+	unsigned long sub = (unsigned char) '';
+	if (cjk_term ||
+		   ((mapped_term /*|| cjk_term*/)
+		    && no_char (mappedtermchar (sub)))) {
+		put_unichar ('>');
+	} else {
+		put_unichar (sub);
+	}
+  }
+  if (with_attr) {
+	clear_attr (attr_uni);
+  }
 }
 
 static
@@ -681,7 +922,7 @@ putshiftmark (marker, utfmarker)
   char * utfmarker;
 {
   if (! mark_mode_active) {
-	mark_on ();
+	set_attr (attr_mark);
 	mark_mode_active = True;
   }
   reverse_on ();
@@ -741,14 +982,14 @@ void
 putCJKtab ()
 {
   unsigned long mark = mappedtermchar (CJK_TAB_marker);
-  mark_on ();
+  set_attr (attr_mark);
   if (no_char (mark)) {
 	put_cjkchar ('.');
 	put_cjkchar ('.');
   } else {
 	put_cjktermchar (mark);
   }
-  mark_off ();
+  clear_attr (attr_mark);
 }
 
 static
@@ -791,14 +1032,14 @@ putCJKret (type)
   }
 
   mark = mappedtermchar (umark);
-  mark_on ();
+  set_attr (attr_mark);
   if (no_char (mark)) {
 	put_cjkchar ('<');
 	put_cjkchar ('<');
   } else {
 	put_cjktermchar (mark);
   }
-  mark_off ();
+  clear_attr (attr_mark);
 }
 
 
@@ -816,8 +1057,8 @@ struct interval {
 };
 
 /*
-struct interval list_Quotation_Mark [] =
-struct interval list_Dash [] =
+static struct interval list_Quotation_Mark [] =
+static struct interval list_Dash [] =
 */
 #include "charprop.t"
 
@@ -881,7 +1122,6 @@ put_cjkcharacter (term, cjkchar, width)
   int width;
 {
   character cjkbytes [5];
-  character * cp;
   unsigned long unichar;
   char encoding_tag = term ? term_encoding_tag : text_encoding_tag;
 
@@ -989,8 +1229,8 @@ put_cjkcharacter (term, cjkchar, width)
 		}
 		return width;
 	} else {
+		character * cp = cjkbytes;
 		trace_put_cjk ("... ", cjkchar);
-		cp = cjkbytes;
 		while (* cp != '\0') {
 			putchar (* cp ++);
 		}
@@ -1049,16 +1289,15 @@ put_cjktermchar (cjkchar)
 }
 
 /**
-   put joined character for separate display to screen
+   put Arabic character to screen, either
+   - isolated, for separated display
+   - ligature, for joined display
  */
 static
 void
-put_unjoined (unichar)
-  unsigned long unichar;
+put_arabic (subst)
+  unsigned long subst;
 {
-  /* prevent terminal ligature by substituting joining character 
-     with its ISOLATED FORM */
-  unsigned long subst = isolated_alef (unichar);
   if ((cjk_term || mapped_term) && no_char (mappedtermchar (subst))) {
 	/* use TATWEEL instead; could use one of
 		U+061F ؟ ARABIC QUESTION MARK
@@ -1113,9 +1352,9 @@ put_unichar (unichar)
   if (mapped_text && no_unichar (unichar)) {
 	indicate_marked ('?');
   } else if (unichar < ' ') {
-	indicate_narrow (unichar + '@');
+	indicate_uni (unichar + '@', False, attr_control);
   } else if (unichar == 0x7F) {
-	indicate_narrow ('?');
+	indicate_uni ('?', False, attr_control);
   } else if (unichar >= 0x80 && unichar <= 0x9F) {
 	if (unichar == 0x9F) {
 		indicate_narrow ('?');
@@ -1188,7 +1427,6 @@ put_unichar (unichar)
   } else {	/* not utf8_screen */
 	unsigned long termchar;
 	int iswide_unichar;
-	int iswide_base;
 	static char * dummyline = "";
 
 	if (cjk_term || mapped_term) {
@@ -1241,6 +1479,12 @@ put_unichar (unichar)
 		} else if (unichar == 0x2713) {
 			/* ✓ */
 			indicate_narrow ('V');
+		} else if (unichar == 0x2717) {
+			/* ✗ */
+			indicate_narrow ('X');
+		} else if (unichar == 0x21AF) {
+			/* ↯ */
+			indicate_narrow ('Z');
 		} else if (unichar == 0x02BB) {
 			/* glottal stop ʻokina */
 			indicate_narrow ('\'');
@@ -1268,6 +1512,7 @@ put_unichar (unichar)
 			}
 		} else {
 			unsigned long base = unichar;
+			int iswide_base;
 			FLAG printable;
 
 #define dont_debug_subst_decompose
@@ -1311,7 +1556,7 @@ put_unichar (unichar)
 					narrow -> narrow
 					narrow -> wide	handled in else branch
 				*/
-				indicate_uni (base);
+				indicate_uni (base, True, attr_uni);
 				if (iswide_base) {
 					/* both (char and subst) are wide */
 					return;
@@ -1434,15 +1679,12 @@ putcharacter (c)
 
 
 /*======================================================================*\
-|*			Text display					*|
+|*			Text line display				*|
 \*======================================================================*/
 
-#define default_script -999
-
-static
-char syntax_marker = syntax_none;
-static
-int script_colour = default_script;
+#define no_script -999
+static char syntax_mask = syntax_none;
+static int current_script_colour;
 
 
 void
@@ -1479,19 +1721,32 @@ clear_lastline ()
 
 static
 void
-disp_syntax (syntax_marker_was, syntax_marker)
-  char syntax_marker_was;
-  char syntax_marker;
+disp_syntax (prev_syntax_mask, syntax_mask)
+  char prev_syntax_mask;
+  char syntax_mask;
 {
-  if (syntax_marker != syntax_marker_was) {
-	if (syntax_marker & syntax_JSP) {
-		dispHTML_jsp ();
-	} else if (syntax_marker & syntax_comment) {
-		dispHTML_comment ();
-	} else if (syntax_marker & syntax_HTML) {
-		dispHTML_code ();
-	} else {
-		dispHTML_off ();
+  if (syntax_mask != prev_syntax_mask) {
+	char new_syntax = syntax_mask & ~ prev_syntax_mask;
+	char old_syntax = prev_syntax_mask & ~ syntax_mask;
+
+	if (old_syntax & syntax_JSP) {
+		clear_attr (attr_jsp);
+	}
+	if (old_syntax & syntax_comment) {
+		clear_attr (attr_comment);
+	}
+	if (old_syntax & syntax_HTML) {
+		clear_attr (attr_html);
+	}
+
+	if (new_syntax & syntax_HTML) {
+		set_attr (attr_html);
+	}
+	if (new_syntax & syntax_comment) {
+		set_attr (attr_comment);
+	}
+	if (new_syntax & syntax_JSP) {
+		set_attr (attr_jsp);
 	}
   }
 }
@@ -1503,21 +1758,19 @@ disp_script (script_colour)
   int script_colour;
 {
   if (use_script_colour) {
-	if (script_colour == default_script) {
-		disp_normal ();
+	if (script_colour == no_script) {
+		clear_colour_attr ();
 	} else {
-		disp_colour (script_colour);
+		set_colour_attr (script_colour);
 	}
   }
 }
 
-static
-struct colour_mapping {
+static struct colour_mapping {
 	char * scriptname;
 	int colour;
 	int colour0;
-} colour_table [] =
-{
+} colour_table [] = {
 #include "colours.t"
 };
 
@@ -1551,7 +1804,7 @@ map_script_colour (scriptname, sc, sc0)
     }
   }
 
-  * sc = default_script;
+  * sc = no_script;
   * sc0 = -1;
 }
 
@@ -1560,11 +1813,11 @@ static
 void
 restore_attr ()
 {
-  if (dim_HTML) {
-	disp_syntax (0, syntax_marker);
+  if (mark_HTML) {
+	disp_syntax (0, syntax_mask);
   }
-  if (script_colour > 0) {
-	disp_script (script_colour);
+  if (current_script_colour > 0) {
+	disp_script (current_script_colour);
   }
 }
 
@@ -1589,6 +1842,7 @@ print_line (ly, line)
 {
   set_cursor (0, ly);
   put_line (ly, line, 0, True, False);
+  line->dirty = False;
 }
 
 /*
@@ -1617,25 +1871,39 @@ put_line (ly, line, offset, clear_line, positioning)
   FLAG separate_combining = True;	/* at line beginning, after TAB */
   FLAG save_combining_mode = combining_mode;
 
-  char syntax_marker_was;
-  char syntax_marker_new;
+  char syntax_mask_old;
+  char syntax_mask_new;
+
+  FLAG sel_highlighting = False;
 
   char * password = NIL_PTR;
 
+  int line_script_colour = no_script;
+
+  clear_attr (attr_all);
+
   if (hide_password) {
 	password = strcontains (textp, "assword");
 	if (password) {
 		password += 7;
-		while (* password == ':' || white_space (* password)) {
+		while (* password == ':' || * password == '=' || white_space (* password)) {
 			password ++;
 		}
 	}
   }
 
-  syntax_marker = line->prev->syntax_marker;
-  syntax_marker_was = syntax_none;
-  syntax_marker_new = syntax_marker;
-  script_colour = default_script;
+  if (mark_HTML) {
+	syntax_mask = line->prev->syntax_mask;
+	syntax_mask_old = syntax_none;
+	syntax_mask_new = syntax_mask;
+	switch (line->return_type) {
+		case lineend_LS:
+		case lineend_PS:
+			line_script_colour = 6;
+			break;
+	}
+  }
+  current_script_colour = line_script_colour;
 
   if (positioning) {
 	offset_start = 0;
@@ -1647,10 +1915,12 @@ put_line (ly, line, offset, clear_line, positioning)
 
 /* Skip all chars as indicated by the offset_start and the shift_count field */
   while (count < offset_start) {
-	syntax_marker_new = syntax_state (syntax_marker_new, textp);
-	if (* textp == '<' || * textp == '>') {
-		syntax_marker_was = syntax_marker;
-		syntax_marker = syntax_marker_new;
+	if (mark_HTML) {
+		syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
+		if (* textp == '<' || * textp == '>') {
+			syntax_mask_old = syntax_mask;
+			syntax_mask = syntax_mask_new;
+		}
 	}
 	advance_char_scr (& textp, & count, line->text);
   }
@@ -1672,10 +1942,12 @@ put_line (ly, line, offset, clear_line, positioning)
 		}
 		if (count == 0) {
 			if (! is_tab (* textp)) {
-				syntax_marker_new = syntax_state (syntax_marker_new, textp);
-				if (* textp == '<' || * textp == '>') {
-					syntax_marker_was = syntax_marker;
-					syntax_marker = syntax_marker_new;
+				if (mark_HTML) {
+					syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
+					if (* textp == '<' || * textp == '>') {
+						syntax_mask_old = syntax_mask;
+						syntax_mask = syntax_mask_new;
+					}
 				}
 				if (iswide (unicodevalue (textp))) {
 					/* skipped character is double-width, 
@@ -1725,6 +1997,16 @@ put_line (ly, line, offset, clear_line, positioning)
 		}
 	}
 
+	if (sel_highlighting && line->sel_end && textp >= line->sel_end) {
+		clear_attr (attr_selected);
+		sel_highlighting = False;
+	} else if (line->sel_begin && ! sel_highlighting 
+			&& textp >= line->sel_begin
+			&& (! line->sel_end || textp < line->sel_end)) {
+		set_attr (attr_selected);
+		sel_highlighting = True;
+	}
+
 	if (* textp == '\n') {
 		charwidth = 1;
 		break;
@@ -1764,14 +2046,14 @@ put_line (ly, line, offset, clear_line, positioning)
 		break;
 	}
 
-	if (dim_HTML) {
-		syntax_marker_new = syntax_state (syntax_marker_new, textp);
+	if (mark_HTML) {
+		syntax_mask_new = syntax_state (syntax_mask_new, syntax_mask_old, textp, line->text);
 		if (* textp == '<' || * textp == '>') {
-			syntax_marker_was = syntax_marker;
-			syntax_marker = syntax_marker_new;
+			syntax_mask_old = syntax_mask;
+			syntax_mask = syntax_mask_new;
 		}
 		if (* textp == '<') {
-			disp_syntax (syntax_marker_was, syntax_marker);
+			disp_syntax (syntax_mask_old, syntax_mask);
 		}
 	}
 
@@ -1831,11 +2113,11 @@ put_line (ly, line, offset, clear_line, positioning)
 		int len;
 
 		put_uniend ();
-		ctrldisp_on ();
+		set_attr (attr_control);
 		for (len = 0; len < charwidth; len ++) {
 			putcharacter ('*');
 		}
-		ctrldisp_off ();
+		clear_attr (attr_control);
 		advance_char (& textp);
 		if (white_space (* textp)) {
 			/*password = NIL_PTR;*/
@@ -1870,9 +2152,9 @@ put_line (ly, line, offset, clear_line, positioning)
 				indicate_narrow (unichar - 0x20);
 			}
 		} else {
-			ctrldisp_on ();
+			set_attr (attr_control);
 			putcharacter (controlchar (unichar));
-			ctrldisp_off ();
+			clear_attr (attr_control);
 		}
 
 		restore_attr ();
@@ -1882,6 +2164,7 @@ put_line (ly, line, offset, clear_line, positioning)
 	} else {
 		FLAG disp_separate_combining = False;
 		FLAG disp_separate_joining = False;
+		FLAG disp_joined = False;
 		char syntax_byte_printed;
 
 		syntax_byte_printed = * textp;
@@ -1892,15 +2175,18 @@ put_line (ly, line, offset, clear_line, positioning)
 
 			follow --;
 
-			last_script_colour = script_colour;
+			last_script_colour = current_script_colour;
 			map_script_colour (script (unichar), 
-				& script_colour, & script_colour0);
-			if (script_colour != last_script_colour) {
+				& current_script_colour, & script_colour0);
+			if (current_script_colour == no_script) {
+				current_script_colour = line_script_colour;
+			}
+			if (current_script_colour != last_script_colour) {
 				if (script_colour0 > 0) {
 					disp_script (script_colour0);
 				}
 				if (colours_256 || script_colour0 <= 0) {
-					disp_script (script_colour);
+					disp_script (current_script_colour);
 				}
 			}
 
@@ -1908,7 +2194,7 @@ put_line (ly, line, offset, clear_line, positioning)
 			    if (combining_screen && ! combining_mode) {
 				/* enforce separated display */
 				disp_separate_combining = True;
-				combiningdisp_on ();
+				set_attr (attr_combining);
 
 				if (iswide (unichar)) {
 					put_unichar (0x3000);
@@ -1944,20 +2230,34 @@ put_line (ly, line, offset, clear_line, positioning)
 			    } else {
 				if (! combining_screen) {
 					disp_separate_combining = True;
-					combiningdisp_on ();
+					set_attr (attr_combining);
 					count += charwidth;
+				} else if (apply_joining
+					&& isjoined (unichar, textp, line->text)) {
+					disp_joined = True;
 				}
 			    }
 			} else {
 				if (iscombining_unichar (unichar)) {
 					disp_separate_combining = True;
-					combiningdisp_on ();
+					set_attr (attr_combining);
 				}
 				count += charwidth;
 			}
 
+			if (disp_joined) {
+				/* apply ligature joining in mined */
+				/* clear previously displayed LAM first */
+				putchar ('');
+				put_unichar (ligature_lam_alef (unichar));
+				/* skip rest of joining char */
+				textp ++;
+				while (follow > 0 && (* textp & 0xC0) == 0x80) {
+					textp ++;
+					follow --;
+				}
 #ifndef mlterm_fixed
-			if (disp_separate_joining) {
+			} else if (disp_separate_joining) {
 				/* prevent terminal ligature by 
 				   substituting joining character 
 				   with its ISOLATED FORM */
@@ -1968,10 +2268,16 @@ put_line (ly, line, offset, clear_line, positioning)
 					textp ++;
 					follow --;
 				}
-			}
-			else
 #endif
-			{
+			} else if ((unichar = disp_subst (unichar))) {
+				indicate_uni (unichar, True, attr_uni);
+				/* skip UTF-8 char */
+				textp ++;
+				while (follow > 0 && (* textp & 0xC0) == 0x80) {
+					textp ++;
+					follow --;
+				}
+			} else {
 				put_char (* textp ++);
 				while (follow > 0 && (* textp & 0xC0) == 0x80) {
 					put_char (* textp ++);
@@ -1979,7 +2285,7 @@ put_line (ly, line, offset, clear_line, positioning)
 				}
 			}
 			if (disp_separate_combining) {
-				combiningdisp_off ();
+				clear_attr (attr_combining);
 				restore_attr ();
 			}
 
@@ -2015,7 +2321,7 @@ put_line (ly, line, offset, clear_line, positioning)
 					    if (combining_screen && ! combining_mode) {
 						/* enforce separated display */
 						disp_separate_combining = True;
-						combiningdisp_on ();
+						set_attr (attr_combining);
 						if (charwidth == 2) {
 							put_unichar (0x3000);
 						} else if (isjoined (unichar, this_textp, line->text)) {
@@ -2026,22 +2332,34 @@ put_line (ly, line, offset, clear_line, positioning)
 					    } else {
 						if (! combining_screen) {
 							disp_separate_combining = True;
-							combiningdisp_on ();
+							set_attr (attr_combining);
+						} else if (apply_joining
+							&& isjoined (unichar, this_textp, line->text)) {
+							disp_joined = True;
 						}
 					    }
 					} else if (iscombining_unichar (unichar)) {
 						disp_separate_combining = True;
-						combiningdisp_on ();
+						set_attr (attr_combining);
 					}
 
-					if (disp_separate_joining) {
-						put_unjoined (unichar);
+					if (disp_joined) {
+						/* apply ligature joining in mined */
+						/* clear previously displayed LAM first */
+						putchar ('');
+						count -= charwidth;
+						put_arabic (ligature_lam_alef (unichar));
+					} else if (disp_separate_joining) {
+						/* prevent terminal ligature 
+						   by substituting joining character 
+						   with its ISOLATED FORM */
+						put_arabic (isolated_alef (unichar));
 					} else {
 						put_cjkcharacter (False, cjkchar, charwidth);
 					}
 
 					if (disp_separate_combining) {
-						combiningdisp_off ();
+						clear_attr (attr_combining);
 						restore_attr ();
 					}
 
@@ -2059,7 +2377,7 @@ put_line (ly, line, offset, clear_line, positioning)
 				    if (combining_screen && ! combining_mode) {
 					/* enforce separated display */
 					disp_separate_combining = True;
-					combiningdisp_on ();
+					set_attr (attr_combining);
 
 					if (isjoined (unichar, textp, line->text)) {
 						disp_separate_joining = True;
@@ -2069,7 +2387,10 @@ put_line (ly, line, offset, clear_line, positioning)
 				    } else {
 					if (! combining_screen) {
 						disp_separate_combining = True;
-						combiningdisp_on ();
+						set_attr (attr_combining);
+					} else if (apply_joining
+						&& isjoined (unichar, textp, line->text)) {
+						disp_joined = True;
 					}
 				    }
 				}
@@ -2078,15 +2399,24 @@ put_line (ly, line, offset, clear_line, positioning)
 			if (mapped_text && unichar == 0xA0 /* nbsp */) {
 				indicate_special ('');
 			} else {
-				if (disp_separate_joining) {
-					put_unjoined (unichar);
+				if (disp_joined) {
+					/* apply ligature joining in mined */
+					/* clear previously displayed LAM first */
+					putchar ('');
+					count -= charwidth;
+					put_arabic (ligature_lam_alef (unichar));
+				} else if (disp_separate_joining) {
+					/* prevent terminal ligature 
+					   by substituting joining character 
+					   with its ISOLATED FORM */
+					put_arabic (isolated_alef (unichar));
 				} else {
 					putcharacter ((character) * textp);
 				}
 			}
 
 			if (disp_separate_combining) {
-				combiningdisp_off ();
+				clear_attr (attr_combining);
 				restore_attr ();
 			}
 
@@ -2094,8 +2424,8 @@ put_line (ly, line, offset, clear_line, positioning)
 			count += charwidth;
 		}
 
-		if (dim_HTML && syntax_byte_printed == '>') {
-			disp_syntax (syntax_marker_was, syntax_marker);
+		if (mark_HTML && syntax_byte_printed == '>') {
+			disp_syntax (syntax_mask_old, syntax_mask);
 		}
 	}
   }
@@ -2106,19 +2436,24 @@ put_line (ly, line, offset, clear_line, positioning)
 
   put_uniend ();
 
-  if (script_colour != default_script) {
-	disp_script (default_script);
+  if (current_script_colour != no_script) {
+	disp_script (no_script);
   }
 
-  if (dim_HTML) {
+  if (mark_HTML) {
 	/* now reset HTML display attribute */
-	disp_syntax (syntax_marker, 0);
+	disp_syntax (syntax_mask, 0);
   }
 
   if (positioning) {
 	/* self-made cursor for terminals (such as xterm)
 	   which have display problems with proportional screen fonts
 	   and their cursor */
+	if (sel_highlighting) {
+		clear_attr (attr_selected);
+		sel_highlighting = False;
+	}
+
 	reverse_on ();
 	if (* textp != '\n') {
 		if (utf8_text) {
@@ -2153,6 +2488,10 @@ put_line (ly, line, offset, clear_line, positioning)
   } else /* (positioning == False) */ {
 	/* If line is longer than XBREAK chars, print the shift_mark */
 	if (count <= XBREAK && * textp != '\n') {
+		if (sel_highlighting) {
+			clear_attr (attr_selected);
+			sel_highlighting = False;
+		}
 		if (count < XBREAK && charwidth == 2) {
 			if (cjk_term) {
 				putmark ('.', NIL_PTR);
@@ -2174,7 +2513,7 @@ put_line (ly, line, offset, clear_line, positioning)
 		}
 	}
 
-	/* Mark end of line if so desired */
+	/* Mark end of line if desired */
 	if (* textp == '\n') {
 	    if (cjk_term) {
 		if (RET_marker >= ' ' && (character) RET_marker <= '@') {
@@ -2187,6 +2526,10 @@ put_line (ly, line, offset, clear_line, positioning)
 				count ++;
 			}
 		}
+		if (sel_highlighting) {
+			clear_attr (attr_selected);
+			sel_highlighting = False;
+		}
 	    } else if (RET_marker != '\0') {
 		if (PARA_marker != '\0' && paradisp
 		 && line->return_type != lineend_LS
@@ -2210,7 +2553,12 @@ put_line (ly, line, offset, clear_line, positioning)
 				}
 			}
 		}
+		if (sel_highlighting) {
+			clear_attr (attr_selected);
+			sel_highlighting = False;
+		}
 
+		/* Fill end of line with markers if desired */
 		count ++;
 		if (marker_defined (RETfill_marker, UTF_RETfill_marker)) {
 			while (count < XBREAK) {
@@ -2233,7 +2581,7 @@ put_line (ly, line, offset, clear_line, positioning)
 #ifdef debug_width
 	static FLAG mark_ret = UNSURE;
 	if (mark_ret == UNSURE) {
-		mark_ret = getenv ("mark_ret") != NIL_PTR;
+		mark_ret = envvar ("mark_ret") != NIL_PTR;
 	}
 	if (mark_ret) {
 		set_cursor (count + 1, ly);
@@ -2250,7 +2598,7 @@ put_line (ly, line, offset, clear_line, positioning)
 			}
 		} else if (can_clear_eol
 			 /* workaround for mlterm bug: */
-			 && ! joining_screen
+			 && mlterm_version == 0
 			  ) {
 			if (count <= XBREAK) {
 				clear_eol ();
@@ -2285,64 +2633,90 @@ set_cursor_xy ()
   }
 }
 
+
+/*======================================================================*\
+|*			Text screen display				*|
+\*======================================================================*/
+
 /*
  * Display line at screen line y_pos if it lies between y_min and y_max.
  * If it is no text line (end of file), clear screen line.
  */
 static
 void
-display_line_at (y_pos, line, y_min, y_max, first)
+display_line_at (y_pos, line, y_min, y_max, first, dirty_only)
   int y_pos;
   int y_min;
   int y_max;
-  register LINE * line;
-  FLAG first;
+  LINE * line;
+  FLAG first;	/* first line being displayed within screen */
+  FLAG dirty_only;
 {
+  /* find line to be displayed */
   line = proceed (line, y_pos - y_min);
+
+  /* skip if dirty_only and not dirty */
+  if (dirty_only && line->dirty == False) {
+	return;
+  }
+
   if (y_pos >= y_min && y_pos <= y_max) {
+	/* clear lines after text, don't postpone this */
 	if (line == tail) {
 		set_cursor (0, y_pos);
 		clear_wholeline ();
-	} else {
-		if (first == False) {
-			if (display_delay >= 0) {
-				flush ();
-			}
-#ifdef msdos
-			if (display_delay > 0) {
-				delay (display_delay);
-			}
-#else
-#ifdef use_usleep
-			if (display_delay > 0) {
-				(void) usleep (1000 * display_delay);
-			}
-#else
-			if (display_delay > 0) {
-				(void) char_ready_within (display_delay, NIL_PTR);
-			}
-#endif
-#endif
-		}
-		print_line (y_pos, line);
+		return;
 	}
+
+	/* postpone actual display unless dirty_only */
+	if (! dirty_only) {
+		line->dirty = True;
+		return;
+	}
+
+/*printf ("display line @ %d\n", y_pos); display_delay = 50;*/
+	if (first == False) {
+		if (display_delay >= 0) {
+			flush ();
+		}
+#ifdef msdos
+		if (display_delay > 0) {
+			delay (display_delay);
+		}
+#else
+# ifdef use_usleep
+		if (display_delay > 0) {
+			(void) usleep (1000 * display_delay);
+		}
+# else
+		if (display_delay > 0) {
+			(void) char_ready_within (display_delay, NIL_PTR);
+		}
+# endif
+#endif
+	}
+	print_line (y_pos, line);
   }
 }
 
 /*
- * Display () shows count + 1 lines on the terminal starting at the given 
- * coordinates. At end of file, the rest of the screen is blanked.
- * When count is negative, a backwards print from 'line' will be done.
+ * display () shows count + 1 lines on the terminal starting at the given 
+   coordinates. At end of file, the rest of the screen is blanked.
+   When count is negative, a backwards print from 'line' will be done.
+ * display_dirty () refreshes only those lines that are marked dirty.
  */
+static
 void
-display (y_coord, line, count, new_pos)
+do_display (y_coord, line, count, new_pos, dirty_only)
   int y_coord;
-  register LINE * line;
-  register int count;
+  LINE * line;
+  int count;
   int new_pos;
+  FLAG dirty_only;
 {
   int y_max = y_coord + count;
   int y_off;
+/*printf ("display %d+%d @ %d (dirty_only %d)\n", y_coord, count, new_pos, dirty_only);*/
 
 /* Find new startline if count is negative */
   if (count < 0) {
@@ -2356,39 +2730,46 @@ display (y_coord, line, count, new_pos)
   menus are now cleared before invoking functions from menu.
 #endif
 
-#define boring_old_top_down_display
-#ifdef old_top_down_display
-/* old code, building the display from top to bottom (how boring): */
-/* with this code, XBREAK must be set to XMAX - 1 */
-/* Print the lines */
-  while (line != tail && count -- >= 0) {
-	print_line (y_coord, line);
-	line = line->next;
-	y_coord ++;
-  }
-
-/* Print the blank lines (if any) */
-  if (loading == False) {
-	while (count -- >= 0) {
-		clear_eol ();
-		putchar ('\n');
-	}
-  }
-#else
-  display_line_at (new_pos, line, y_coord, y_max, True);
+  display_line_at (new_pos, line, y_coord, y_max, True, dirty_only);
   y_off = 0;
   while (y_off <= count) {
 	y_off ++;
 	if (winchg) {
 		return;
 	}
-	display_line_at (new_pos - y_off, line, y_coord, y_max, False);
+	display_line_at (new_pos - y_off, line, y_coord, y_max, False, dirty_only);
 	if (winchg) {
 		return;
 	}
-	display_line_at (new_pos + y_off, line, y_coord, y_max, False);
+	display_line_at (new_pos + y_off, line, y_coord, y_max, False, dirty_only);
   }
-#endif
+}
+
+static
+void
+display_dirty (y_coord, line, count, new_pos)
+  int y_coord;
+  LINE * line;
+  int count;
+  int new_pos;
+{
+  do_display (y_coord, line, count, new_pos, True);
+}
+
+void
+display_flush ()
+{
+  display_dirty (0, top_line, last_y, y);
+}
+
+void
+display (y_coord, line, count, new_pos)
+  int y_coord;
+  LINE * line;
+  int count;
+  int new_pos;
+{
+  do_display (y_coord, line, count, new_pos, False);
 }
 
 
