@@ -7,6 +7,11 @@
 #include "io.h"
 #include "termprop.h"
 
+#ifndef VAXC
+#include <dirent.h>
+#include <sys/stat.h>
+#endif
+
 
 FLAG char_on_status_line = False; /* is output active on status line ? */
 FLAG input_active = False;
@@ -19,7 +24,7 @@ FLAG input_active = False;
 /* bottom status contents, kept static for redraw */
 static char * bottom_inbuf;
 static FLAG bottom_attrib;
-static char bottom_status_buf [maxLINE_LEN];
+static char bottom_status_buf [2 * maxPROMPTlen + maxFILENAMElen];
 static int bottom_begin_fmt;
 static FLAG bottom_utf8_text;
 
@@ -253,6 +258,12 @@ print_unichar (unichar, width, char_begin, line_begin)
 		/*	disp_separate_combining = True;	*/
 		/*	combiningdisp_on ();		*/
 			put_unichar (unichar);
+		} else if (apply_joining
+			&& isjoined (unichar, char_begin, line_begin)) {
+			/* apply ligature joining in mined */
+			/* clear previously displayed LAM first */
+			putchar ('');
+			put_unichar (ligature_lam_alef (unichar));
 		} else {
 			put_unichar (unichar);
 		}
@@ -379,13 +390,14 @@ print_string (text)
 \*======================================================================*/
 
 /*
- * promptyn reads in a 'y' or 'n' character.
+ * prompt reads in a 'y' or 'n' or other character.
  */
 character
-promptyn ()
+prompt (term)
+  char * term;
 {
-  register character c;
-  while ((c = readcharacter ()) != 'y' && c != 'n' && c != '\033' && ! quit) {
+  character c;
+  while ((c = readcharacter ()) != '\033' && ! strchr (term, c) && ! quit) {
 	ring_bell ();
 	flush ();
   }
@@ -403,8 +415,11 @@ promptyn ()
 void
 swallow_dummy_quit_char ()
 {
-  (void) readcharacter (); /* Swallow away a quit character delivered by QUIT */
-/* Not needed because this character is ignored by being the CANCEL command */
+  /* Swallow away a quit character delivered by QUIT;
+     not needed because this character is 
+     ignored by being the CANCEL command
+   */
+  (void) _readchar_nokeymap ();
 }
 #endif
 
@@ -500,7 +515,7 @@ input_byte (inbuf, ptr, c)
   char * ptr;
   character c;
 {
-  if (c > '\0' && (ptr - inbuf) < maxLINE_LEN) {
+  if (c > '\0' && (ptr - inbuf) < maxPROMPTlen) {
 	if (mapped_text && ! combining_mode && encoding_has_combining ()) {
 		print_mapped (c, ptr, inbuf);
 	} else {
@@ -546,12 +561,11 @@ input_cjk (inbuf, ptr, c)
 {
   character cjkbytes [5];
   int len;
-  int scrlen;
   character * cp;
 
 	len = cjkencode (c, cjkbytes);
-	if (len > 0 && (ptr - inbuf - 1 + len) < maxLINE_LEN) {
-		scrlen = cjkscrwidth (c, ptr, inbuf);
+	if (len > 0 && (ptr - inbuf - 1 + len) < maxPROMPTlen) {
+		int scrlen = cjkscrwidth (c, ptr, inbuf);
 
 		char_on_status_line = True;
 
@@ -631,7 +645,7 @@ input_unicode (inbuf, ptr, unichar)
 	if (unichar < 0x80) {
 	    return input_byte (inbuf, ptr, unichar);
 	} else if (unichar < 0x800) {
-	    if ((ptr - inbuf) + 2 > maxLINE_LEN) {
+	    if ((ptr - inbuf) + 2 > maxPROMPTlen) {
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
@@ -640,7 +654,7 @@ input_unicode (inbuf, ptr, unichar)
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x10000) {
-	    if ((ptr - inbuf) + 3 > maxLINE_LEN) {
+	    if ((ptr - inbuf) + 3 > maxPROMPTlen) {
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
@@ -650,7 +664,7 @@ input_unicode (inbuf, ptr, unichar)
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x200000) {
-	    if ((ptr - inbuf) + 4 > maxLINE_LEN) {
+	    if ((ptr - inbuf) + 4 > maxPROMPTlen) {
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
@@ -661,7 +675,7 @@ input_unicode (inbuf, ptr, unichar)
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x4000000) {
-	    if ((ptr - inbuf) + 5 > maxLINE_LEN) {
+	    if ((ptr - inbuf) + 5 > maxPROMPTlen) {
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
@@ -673,7 +687,7 @@ input_unicode (inbuf, ptr, unichar)
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x80000000) {
-	    if ((ptr - inbuf) + 6 > maxLINE_LEN) {
+	    if ((ptr - inbuf) + 6 > maxPROMPTlen) {
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
@@ -988,7 +1002,10 @@ input (inbuf, clearfl, term_input)
 	}
      } else if (command (c) == CTRLINS) {
 	c = readcharacter_unicode ();
-	if (c == FUNcmd) {
+	if (command (c) == COMPOSE && keyshift >= '@' && keyshift <= '_') {
+		/* allow input of control char, esp ^^ */
+		ptr = input_character (inbuf, ptr, keyshift & 0x1F);
+	} else if (c == FUNcmd) {
 		struct prefixspec * ps;
 		keyshift |= ctrl_mask;
 		ps = lookup_prefix (keyproc, keyshift);
@@ -1013,7 +1030,7 @@ input (inbuf, clearfl, term_input)
 		mnemo = (char *) ptr;
 		diagdisp_off ();
 	} else if (c > ' ' && c != 0x7F) {
-		c = compose (c, readcharacter_unicode ());
+		c = compose_chars (c, readcharacter_unicode ());
 		if (no_char (c)) {
 			ring_bell ();
 		} else {
@@ -1022,7 +1039,12 @@ input (inbuf, clearfl, term_input)
 	} else {
 		ptr = input_character (inbuf, ptr, c);
 	}
-     } else if (c == '\033' || command (c) == MOUSEescape) {
+     } else if (c == '\033'
+		|| (command (c) == MOUSEfunction
+			/* workaround for mintty shift-paste */
+			&& mouse_button != releasebutton
+			/* consider fix_mouse_release_event () */
+			&& mouse_button != movebutton)) {
 	quit = True;
      } else if (c == '\n' || c == '\r') {
 	/* if inbuf is empty clear status line */
@@ -1050,6 +1072,7 @@ input (inbuf, clearfl, term_input)
 status_line (str1, str2) is (void) bottom_line (True, (str1), (str2), NIL_PTR, False, "")
 status_msg (str)	 is status_line (str, NIL_PTR)
 status_beg (str)	 is (void) bottom_line (True, (str), NIL_PTR, NIL_PTR, True, "")
+status_fmt2(str1, str2)	 is (void) bottom_line (VALID, str1, str2, NIL_PTR, False, "")
 error (str1)		 is (void) bottom_line (True, (str1), NIL_PTR, NIL_PTR, False, "")
 error2 (str1, str2)	 is (void) bottom_line (True, (str1), (str2), NIL_PTR, False, "")
 clear_status ()		 is (void) bottom_line (False, NIL_PTR, NIL_PTR, NIL_PTR, False, "")
@@ -1057,19 +1080,38 @@ get_string (str1, str2, fl, term_chars)
 			 is bottom_line (True, (str1), NIL_PTR, (str2), fl, term_chars)
  */
 
+static
 void
-status_uni (msg)
-  char * msg;
+status2_uni (msg1, msg2)
+  char * msg1;
+  char * msg2;
 {
   FLAG save_utf8_text = utf8_text;
   utf8_text = True;
   /* don't need to save and set utf16_file = False; */
 
-  bottom_line (VALID, msg, NIL_PTR, NIL_PTR, False, "");
+  bottom_line (VALID, msg1, msg2, NIL_PTR, False, "");
 
   utf8_text = save_utf8_text;
 }
 
+void
+status_uni (msg)
+  char * msg;
+{
+  status2_uni (msg, NIL_PTR);
+}
+
+character
+status2_prompt (term, msg1, msg2)
+  char * term;
+  char * msg1;
+  char * msg2;
+{
+  (void) bottom_line (VALID, msg1, msg2, NIL_PTR, False, term);
+  return prompt (term);
+}
+
 /**
    get_string_uni prompts a string like get_string_nokeymap
    but uses Unicode input
@@ -1153,6 +1195,10 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
 {
   int ret = FINE;
 
+  if (debug_mined && (s1 || s2)) {
+	debuglog ("prompt", s1, s2);
+  }
+
   if (inbuf != NIL_PTR) {
 	* inbuf = '\0';
   }
@@ -1160,6 +1206,7 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   if (pagewrapped) {
 	bottom_status_buf [0] = '\0';
 	RD ();
+	display_flush ();
 	pagewrapped = False;
   }
 
@@ -1238,6 +1285,8 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
 	set_cursor (0, YMAX);
   } else if (statfl) {
 	diagdisp_on ();
+  } else if (term_input && * term_input) {
+	/* leave cursor for prompt () */
   } else {
 	set_cursor_xy ();	/* Set cursor back to old position */
   }
@@ -1249,31 +1298,9 @@ bottom_line (attrib, s1, s2, inbuf, statfl, term_input)
   return ret;
 }
 
-/*
- * Get_file () reads a filename from the terminal.
- */
-int
-get_filename (message, file)
-  char * message;
-  char * file;
-{
-  int ret = get_string (message, file, True, "");
-#ifndef msdos
-  char * filei;
-  char file1 [maxLINE_LEN];
-
-  if (file [0] == '~' && (file [1] == '/' || file [1] == '\0')) {
-	filei = file; filei ++;
-	build_string (file1, "%s%s", envvar ("HOME"), filei);
-	strcpy (file, file1);
-  }
-#endif
-  return ret;
-}
-
 
 /*======================================================================*\
-|*				Number input				*|
+|*			Number input					*|
 \*======================================================================*/
 
 /*
@@ -1462,6 +1489,474 @@ prompt_num_char (result, maxvalue)
 }
 
 
+/*======================================================================*\
+|*			File chooser					*|
+\*======================================================================*/
+
+FLAG sort_by_extension = False;
+FLAG sort_dirs_first = False;
+
+#ifdef VAXC
+
+/*
+ * get_filename () prompts for a file name (on the prompt line)
+ */
+int
+get_filename (message, file, directory_chooser)
+  char * message;
+  char * file;
+  FLAG directory_chooser;
+{
+  char file0 [maxPROMPTlen];
+  int ret = get_string (message, file0, True, "");
+  char * filep = file0;
+
+  strcpy (file, "");
+
+#ifndef msdos
+  if (file0 [0] == '~' && (file0 [1] == '/' || file0 [1] == '\0')) {
+	strappend (file, gethomedir (), maxFILENAMElen);
+	filep ++;
+  }
+#endif
+
+  strappend (file, filep, maxFILENAMElen);
+  return ret;
+}
+
+#else
+
+#define dont_debug_file_chooser
+#define dont_debug_wildcard
+
+
+struct treenode {
+	struct treenode * left;
+	struct treenode * right;
+	char * s1;
+	char * s2;
+};
+
+static struct treenode * dir_root = 0;
+
+static
+char *
+getsuffix (fn)
+  char * fn;
+{
+  while (* fn == '.') {
+	fn ++;
+  }
+  return unnull (strrchr (fn, '.'));
+}
+
+static
+int
+fncompare (s1, s2)
+  char * s1;
+  char * s2;
+{
+  if (sort_by_extension) {
+	char * suf1 = getsuffix (s1);
+	char * suf2 = getsuffix (s2);
+	int cmp = stringcompare_ (suf1, suf2);
+	if (cmp != 0) {
+		return cmp;
+	} else {
+		return stringcompare_ (s1, s2);
+	}
+  } else {
+	return stringcompare_ (s1, s2);
+  }
+}
+
+static
+void
+tree_insert (nodepoi, s1, s2)
+  struct treenode * * nodepoi;
+  char * s1;
+  char * s2;
+{
+  if (* nodepoi) {
+	int cmp = fncompare (s1, (* nodepoi)->s1);
+	if (cmp < 0) {
+		tree_insert (& (* nodepoi)->left, s1, s2);
+	} else if (cmp > 0) {
+		tree_insert (& (* nodepoi)->right, s1, s2);
+	} else {
+		/* duplicate entry? */
+	}
+  } else {
+	* nodepoi = alloc (sizeof (struct treenode));
+	(* nodepoi)->left = 0;
+	(* nodepoi)->right = 0;
+	(* nodepoi)->s1 = s1;
+	(* nodepoi)->s2 = s2;
+  }
+}
+
+static menuitemtype * filemenu;
+static int itemcount;
+
+static
+void
+tree_traverse (node, dirs)
+  struct treenode * node;
+  FLAG dirs;
+{
+  if (node) {
+	tree_traverse (node->left, dirs);
+	if (dirs == NOT_VALID ||
+	    (dirs == False) == ! node->s2) {
+		fill_menuitem (& filemenu [itemcount ++], node->s1, node->s2);
+	}
+	tree_traverse (node->right, dirs);
+  }
+}
+
+static
+void
+tree_free (node)
+  struct treenode * node;
+{
+  if (node) {
+	tree_free (node->left);
+	tree_free (node->right);
+	free_space (node->s1);
+	free_space (node);
+  }
+}
+
+
+static
+int
+isin (utf, set)
+  char * utf;
+  char * set;
+{
+  unsigned long uni = utf8value (utf);
+
+  while (1) {
+	unsigned long elem = utf8value (set);
+	if (uni == elem) {
+		return 1;
+	}
+	set = nextutf8 (set);
+	if (* set == '-') {
+		unsigned long elem2;
+		set ++;
+		elem2 = utf8value (set);
+		if (uni > elem && uni <= elem2) {
+			return 1;
+		}
+		if (elem2 && elem2 != ']') {
+			set = nextutf8 (set);
+		}
+	}
+	if (* set == ']') {
+		return 0;
+	}
+  }
+}
+
+static
+int
+wildcard (pat, s)
+  char * pat;
+  char * s;
+{
+#ifdef debug_wildcard
+  indent (level);
+  printf ("wildcard (%s, %s)\n", pat, s);
+#endif
+
+  if (! * pat) {
+	return ! * s;
+  } else if (* pat == '?') {
+	return * s && wildcard (pat + 1, nextutf8 (s));
+  } else if (* pat == '*') {
+	return (* s && wildcard (pat, nextutf8 (s)))
+		|| wildcard (pat + 1, s);
+  } else if (* pat == '[') {
+	char * endset = strchr (pat, ']');
+	if (endset) {
+		pat ++;
+		if (* pat == '^' || * pat == '!') {
+			pat ++;
+			return ! isin (s, pat)
+				&& wildcard (endset + 1, nextutf8 (s));
+		} else {
+			return isin (s, pat)
+				&& wildcard (endset + 1, nextutf8 (s));
+		}
+	} else {
+		/* invalid pattern */
+		return 0;
+	}
+  } else {
+#ifdef case_sensitive
+	return * pat == * s	/* transparently compare UTF-8 bytes */
+		&& wildcard (pat + 1, s + 1);
+#else
+	unsigned long c1 = case_convert (utf8value (pat), -1);
+	unsigned long c2 = case_convert (utf8value (s), -1);
+	return c1 == c2
+		&& wildcard (nextutf8 (pat), nextutf8 (s));
+#endif
+  }
+}
+
+static
+char *
+pathlast (path)
+  char * path;
+{
+  char * bn = getbasename (path);
+  if (bn == path) {
+	return NIL_PTR;
+  } else {
+	return bn - 1;
+  }
+}
+
+#ifndef S_ISDIR
+#define S_ISDIR(mode)	(mode & S_IFMT) == S_IFDIR
+#endif
+
+/*
+ * get_filename () opens a file chooser dialog to choose a file name
+ */
+int
+get_filename (message, file, directory_chooser)
+  char * message;
+  char * file;
+  FLAG directory_chooser;
+{
+  char dirname [maxFILENAMElen];
+  char * filter = NIL_PTR;
+  char dirpath [maxFILENAMElen];
+  FLAG continue_chooser = True;
+  int ret = ERRORS;
+
+  strcpy (dirname, ".");
+
+  while (continue_chooser) {
+	struct dirent * direntry;
+	int filecount = 0;
+	int fi;
+	DIR * dir = opendir (dirname);
+	if (! dir) {
+		error2 ("Cannot open directory: ", dirname);
+		return ERRORS;
+	}
+
+	dir_root = 0;
+	while ((direntry = readdir (dir))) {
+		if (streq (direntry->d_name, ".")
+		    /* && ! directory_chooser */
+		   ) {
+			/* ignore */
+			/* also for change dir, if absolute cwd is listed */
+		} else if (! filter || wildcard (filter, direntry->d_name)) {
+			char * dirflag = NIL_PTR;
+#ifdef bogus_DIRENT_HAVE_D_TYPE
+			if (direntry->d_type == DT_DIR) {
+				dirflag = "directory";
+			}
+#else
+			struct stat fstat_buf;
+			char pathname [maxFILENAMElen];
+			strcpy (pathname, dirname);
+			if (dirname [strlen (dirname) - 1] != '/'
+#ifdef pc
+			 && dirname [strlen (dirname) - 1] != '\\'
+#endif
+			   ) {
+				strappend (pathname, "/", maxFILENAMElen);
+			}
+			strappend (pathname, direntry->d_name, maxFILENAMElen);
+			if (stat (pathname, & fstat_buf) < 0) {
+				/* No such file or directory ... */
+			} else if (S_ISDIR (fstat_buf.st_mode)) {
+				dirflag = "directory";
+			}
+#endif
+			if (! directory_chooser || dirflag) {
+				tree_insert (& dir_root, dupstr (direntry->d_name), dirflag);
+				filecount ++;
+			}
+		}
+	}
+	closedir (dir);
+	/* reset wildcard filter */
+	filter = NIL_PTR;
+
+	/* adjust # entries */
+	filecount += 1;
+	/* allocate menu structure */
+	filemenu = alloc (filecount * sizeof (menuitemtype));
+	itemcount = 0;
+	/* compose pathname of selected directory for display */
+	if (is_absolute_path (dirname)) {
+		strcpy (dirpath, dirname);
+	} else {
+		char * dnpoi;
+		(void) get_cwd (dirpath);
+		dnpoi = dirname;
+		while (strisprefix ("..", dnpoi) &&
+			(dnpoi [2] == '\0'
+			 || dnpoi [2] == '/'
+#ifdef pc
+			 || dnpoi [2] == '\\'
+#endif
+			)
+		      ) {
+			char * lastslash = pathlast (dirpath);
+			if (lastslash) {
+				* lastslash = 0;
+			}
+			dnpoi += 2;
+			if (* dnpoi) {
+				dnpoi ++;
+			}
+		}
+		if (* dnpoi && ! streq (dnpoi, ".")) {
+			strappend (dirpath, "/", maxFILENAMElen);
+			strappend (dirpath, dnpoi, maxFILENAMElen);
+		}
+		if (! * dirpath) {
+			strcat (dirpath, "/");
+		}
+	}
+	/* pre-fill current working directory */
+	fill_menuitem (& filemenu [itemcount ++], dirpath, "directory");
+	/* fill menu with dir/file names */
+	if (sort_dirs_first) {
+		tree_traverse (dir_root, True);
+		tree_traverse (dir_root, False);
+	} else {
+		tree_traverse (dir_root, NOT_VALID);
+	}
+
+	hop_flag = 0;
+	fi = popup_menu (filemenu, filecount, 2, 2, message, True, False, "*");
+
+	if (fi < 0) {
+		ret = ERRORS;
+		continue_chooser = False;
+	} else if (streq (filemenu [fi].itemname, "~")) {
+		strcpy (dirname, "");
+		strappend (dirname, gethomedir (), maxFILENAMElen);
+	} else {
+	    char * chosenname = filemenu [fi].itemname;
+	    char * dirtag = filemenu [fi].hopitemname;
+	    char expandedname [maxFILENAMElen];
+	    if (chosenname [0] == '~'
+		&& (chosenname [1] == '/' || chosenname [1] == '\0')) {
+		strcpy (expandedname, "");
+		strappend (expandedname, gethomedir (), maxFILENAMElen);
+		chosenname ++;
+		strappend (expandedname, chosenname, maxFILENAMElen);
+		chosenname = expandedname;
+	    }
+
+#ifdef debug_file_chooser
+	    printf ("%s -> %s (%s)", dirname, chosenname, dirtag);
+#endif
+	    if (streq (dirtag, "*")) {
+		/* set wildcard filter */
+		filter = chosenname;
+	    } else {
+		/* compose changed directory/file name */
+		if (streq (chosenname, "..")) {
+			char * lastslash = pathlast (dirname);
+			if ((lastslash && streq (lastslash + 1, ".."))
+			    || streq (dirname, "..")) {
+				strappend (dirname, "/..", maxFILENAMElen);
+			} else if (lastslash) {
+				if (lastslash == dirname) {
+					lastslash ++;
+				}
+				* lastslash = 0;
+			} else if (streq (dirname, ".")) {
+				strcpy (dirname, "..");
+			} else {
+				strcpy (dirname, ".");
+			}
+#ifdef pc_better_dont
+			if (dirname [1] == ':' && dirname [2] == '\0') {
+				strcat (dirname, "\\");
+			}
+#endif
+		} else if (streq (chosenname, ".")) {
+			/* no change */
+		} else {
+			if (streq (dirname, ".") || is_absolute_path (chosenname)) {
+				strcpy (dirname, "");
+			} else {
+				if (dirname [strlen (dirname) - 1] != '/'
+#ifdef pc
+				 && dirname [strlen (dirname) - 1] != '\\'
+#endif
+				   ) {
+					strappend (dirname, "/", maxFILENAMElen);
+				}
+			}
+			strappend (dirname, chosenname, maxFILENAMElen);
+		}
+
+#ifdef msdos
+		if (chosenname [0] && chosenname [1] == ':' && ! chosenname [2]) {
+			/* MSDOS shortcut for change disk */
+			strcpy (file, chosenname);
+			ret = FINE;
+			continue_chooser = False;
+		}
+		else
+#endif
+		if (directory_chooser) {
+			if (streq (dirtag, "cd")) {
+				/* navigate into directory after TAB */
+			} else {
+				/* exit with selected directory name */
+				strcpy (file, dirname);
+				ret = FINE;
+				continue_chooser = False;
+			}
+		} else if (streq (dirtag, "directory")) {
+			/* navigate into selected directory */
+		} else {
+			/* check whether selected name 
+			   is actually a directory
+			*/
+			struct stat fstat_buf;
+			if (stat (dirname, & fstat_buf) == 0
+			 && S_ISDIR (fstat_buf.st_mode)) {
+				/* navigate into directory */
+			} else {
+				/* exit with selected file name */
+				strcpy (file, dirname);
+				ret = FINE;
+				continue_chooser = False;
+			}
+		}
+	    }
+#ifdef debug_file_chooser
+	    printf (": %s\n", dirname);
+#endif
+	}
+
+	free_space (filemenu);
+	tree_free (dir_root);
+	dir_root = 0;
+  }
+
+  return ret;
+}
+
+#endif
+
+
 /*======================================================================*\
 |*				End					*|
 \*======================================================================*/
