@@ -68,7 +68,7 @@ static FLAG wrapped_search;		/* temporary flag of wrapped search */
 static REGEX * lastprogram = NIL_REG;
 static FLAG lastmethod = NOT_VALID;	/* FORWARD, REVERSE */
 
-static char prevexpr [maxLINE_LEN];	/* Buffer for previous expr. */
+static char prevexpr [maxPROMPTlen];	/* Buffer for previous expr. */
 static FLAG prevmethod = NOT_VALID;	/* FORWARD, REVERSE */
 
 
@@ -94,7 +94,7 @@ static FLAG prevmethod = NOT_VALID;	/* FORWARD, REVERSE */
  *	   that expression.
  */
 
-static char typed_expression [maxLINE_LEN];	/* Holds previous search expression */
+static char typed_expression [maxPROMPTlen];	/* Holds previous search expression */
 
 /*
  * SIDFW searches forward for the current identifier.
@@ -225,7 +225,7 @@ RSreverse ()
  * The save flag indicates whether the expression should be appended at the
  * message pointer.
  */
-static char exp_buf [maxLINE_LEN];	/* Buffer for new expr. */
+static char exp_buf [maxPROMPTlen];	/* Buffer for new expr. */
 static REGEX program;			/* Program of expression */
 
 static
@@ -297,8 +297,8 @@ change (message, global, confirm)
   FLAG global;
   FLAG confirm;
 {
-  char mess_buf [maxLINE_LEN];		/* Buffer to hold message */
-  char replacement [maxLINE_LEN];	/* Buffer to hold subst. pattern */
+  char mess_buf [maxPROMPTlen];		/* Buffer to hold message */
+  char replacement [maxPROMPTlen];	/* Buffer to hold subst. pattern */
   REGEX * program;			/* Program resulting from compilation */
   LINE * line = cur_line;
   char * searchfrom = cur_text;
@@ -308,12 +308,10 @@ change (message, global, confirm)
   long lines = 0L;		/* Nr of lines on which subs occurred */
   int ly = y;			/* Index to check if line is on screen */
   int previousy = y;
-  char c;
   FLAG quit_change;
   int lines_changed = 0;
 
-  if (viewonly) {
-	viewonlyerr ();
+  if (dont_modify ()) {
 	return;
   }
 
@@ -359,13 +357,17 @@ change (message, global, confirm)
 	    /* Repeat sub. on this line as long as we find a match */
 	    do {
 		if (confirm) {
+			char c;
 			ly = find_y (line);
 			textp = program->start_ptr;
 			move_address (program->start_ptr, ly);
-			status_msg ("Replace ? (y/n)");
-			c = promptyn ();
+			display_flush ();
+			set_cursor_xy ();
+			status_uni ("Replace ? (y/n/ESC)");
+			c = prompt ("yn");
 			clear_status ();
 			if (c == 'y') {
+			    int prev_last_y = last_y;
 			    subs ++;	/* Increment # substitutions made */
 			    textp = substitute (line, & line, 
 					program, & old_lines_deleted, 
@@ -374,7 +376,11 @@ change (message, global, confirm)
 				/* reset bot_line and last_y */
 				reset (top_line, y);
 				/* update display */
-				display (y, cur_line, last_y - y, y);
+				if (prev_last_y > last_y) {
+					display (y, cur_line, prev_last_y - y, y);
+				} else {
+					display (y, cur_line, last_y - y, y);
+				}
 			    } else {
 				print_line (ly, line);
 			    }
@@ -454,7 +460,7 @@ change (message, global, confirm)
   if (subs == 0L && quit == False) {
 	error ("Pattern not found");
   } else if (lines >= REPORT_CHANGED_LINES || quit) {
-	build_string (mess_buf, "%s %ld substitutions on %ld lines",
+	build_string (mess_buf, "%s%ld substitutions on %ld lines",
 		quit_change ? "(Replace aborted) " : "", subs, lines);
 	status_msg (mess_buf);
   } else if (global == NOT_VALID && subs >= REPORT_CHANGED_LINES) {
@@ -567,7 +573,7 @@ substitute (line, line_poi,
 
 /* Check if substituting in more than 1 line; delete lines then */
   if (program->end_line != program->start_line) {
-	(void) delete_text (line->next, line->next->text, 
+	(void) delete_text_only (line->next, line->next->text, 
 		     program->end_line->next, program->end_line->next->text);
 	* old_lines_deleted = 9;
   } else {
@@ -616,7 +622,7 @@ substitute (line, line_poi,
 	}
 	repl_len = (int) (repl_end - repl_start);
 
-	line = line_insert (line, repl_start, repl_len, line->return_type);
+	line = line_insert_after (line, repl_start, repl_len, line->return_type);
 	if (line == NIL_LINE) {
 		ring_bell ();
 		error ("Substitution failed: cannot allocate more lines");
@@ -697,7 +703,7 @@ search (message, method)
   FLAG method;
 {
   register REGEX * program;
-  char prevexp_buf [maxLINE_LEN];	/* Buffer for previous expr. */
+  char prevexp_buf [maxPROMPTlen];	/* Buffer for previous expr. */
 
   if (lastmethod != NOT_VALID) {
 	copy_string (prevexp_buf, exp_buf);
@@ -777,8 +783,6 @@ search_corresponding (expr, method, corresponding)
 {
   register REGEX * program;
   LINE * match_line;
-  int matched;
-  int nestingweight;
   int bracount = 0;
   LINE * old_line = cur_line;
   char * old_text = cur_text;
@@ -798,6 +802,8 @@ search_corresponding (expr, method, corresponding)
 #endif
   wrapped_search = False;
   do {
+	int matched;
+	int nestingweight;
 	do {
 		match_line = search_it (program, method);
 		if (match_line == NIL_LINE) {
@@ -949,6 +955,10 @@ do_search (program, method)
 	return;
   }
 
+  if (! visselect_keeponsearch) {
+	clear_highlight_selection ();
+  }
+
   match_line = search_it (program, method);
   if (match_line == NIL_LINE) {
 	return;
@@ -992,42 +1002,6 @@ static char * too_long = "Regular expression too long";
 #define reg_error(str)	program->status = REG_ERROR, \
 					program->result.err_mess = (str)
 
-/*
- * Bcopy copies 'bytes' bytes from the 'from' address into the 'to' address.
- */
-#ifdef vms
-#define defbcopy
-#endif
-#ifdef msdos
-# ifdef __TURBOC__
-#define defbcopy
-# endif
-#endif
-
-#ifdef defbcopy	/* otherwise also in standard library */
-static
-void
-bcopy (from, to, bytes)
-  register char * from;
-  register char * to;
-  register int bytes;
-{
-  while (bytes --) {
-	* to ++ = * from ++;
-  }
-}
-#else
-#ifdef sysV
-#define bcopy(from, to, len)	memcpy (to, from, len)
-#else
-# ifdef msdos
-#define bcopy(from, to, len)	memcpy (to, from, len)
-# else
-extern void bcopy ();
-# endif
-#endif
-#endif
-
 /*
  * Finished () is called when everything went right during compilation. It
  * allocates space for the expression, and copies the expression buffer into
@@ -1047,7 +1021,7 @@ finished (program, last_exp)
 	error ("Cannot allocate memory for search expression");
 	return ERRORS;
   } else {	/* Copy expression. (expression consists of ints!) */
-	bcopy (exp_buffer, program->result.expression, length);
+	memcpy (program->result.expression, exp_buffer, length);
 	return FINE;
   }
 }
@@ -1408,7 +1382,6 @@ match (program, string, method)
   register FLAG method;
 {
   register LINE * line = cur_line;
-  character old_charbyte;		/* For saving chars */
   char * next = (char *) string;
 
 /* Corrupted program */
@@ -1429,8 +1402,8 @@ match (program, string, method)
 	}
   }
   if (method == REVERSE && ! (program->status & END_LINE)) {
-	old_charbyte = * string;	/* Save char and */
-	* string = '\n';	/* Assign '\n' for line_check */
+	character old_charbyte = * string;	/* Save char and */
+	* string = '\n';		/* Assign '\n' for line_check */
 	if (line_check (program, line->text, line, method) == MATCH) {
 		* string = old_charbyte; /* Restore char */
 		trace_search (("-> match 2 found\n"));
@@ -1478,7 +1451,6 @@ line_check (program, string, from_line, method)
   FLAG method;
 {
   char * textp = string;
-  int ret;
 
   trace_search (("line_check [%d] %s", string - from_line->text, string));
   trace_indent ();
@@ -1490,7 +1462,7 @@ line_check (program, string, from_line, method)
 /* If the match must be anchored, just check the string. */
   if (program->status & BEGIN_LINE) {
 	if (string == from_line->text) {
-		ret = check_string (program, string, from_line, NIL_INT);
+		int ret = check_string (program, string, from_line, NIL_INT);
 		trace_undent ();
 		return ret;
 	} else {
@@ -1807,8 +1779,6 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 {
   character * spoi;
   FLAG charmatch;
-  character cjkbuf [2];
-  int cjklen;
 
   if (* s == '\0' || (* s == '\n' && (opcode & NEGATE))) {
 	/* End of string, never matches */
@@ -1844,6 +1814,8 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 	}
   } else if (cjk_text) {
 	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
+		character cjkbuf [2];
+		int cjklen;
 		cjkbuf [0] = * exprlist & LOW_BYTE;
 		if (multichar (cjkbuf [0])) {
 			cjkbuf [1] = * (exprlist + 1) & LOW_BYTE;
