@@ -16,6 +16,7 @@ FLAG utf8_input = False;	/* keyboard input in UTF-8 mode ? */
 FLAG combining_screen = False;	/* combining character terminal ? */
 FLAG bidi_screen = False;	/* UTF-8 bidi terminal ? */
 FLAG joining_screen = False;	/* UTF-8 terminal with Arabic LAM/ALEF joining ? */
+FLAG halfjoining_screen = False;	/* Putty/mintty joining + spare space ? */
 
 FLAG mapped_term = False;	/* terminal in mapped 8-bit mode ? */
 
@@ -48,6 +49,7 @@ int gnome_terminal_version = 0;
 int rxvt_version = 0;
 int konsole_version = 0;
 int mintty_version = 0;
+int mlterm_version = 0;
 int screen_version = 0;
 
 /* terminal features with respect to non-BMP characters (>= 0x10000) */
@@ -69,12 +71,15 @@ static struct name_table {
 	{U400, "Unicode 4.0"},
 	{U410, "Unicode 4.1"},
 	{U500, "Unicode 5.0"},
+	{U510, "Unicode 5.1"},
+	{U520, "Unicode 5.2"},
+	{U600, "Unicode 6.0"},
 	{0, 0}
 };
 
-
+/* return name of Unicode version applied by terminal */
 char *
-term_encoding_version_name (v)
+term_Unicode_version_name (v)
   int v;
 {
   struct name_table * t = Unicode_version_names;
