@@ -22,6 +22,7 @@ extern FLAG utf8_input;		/* keyboard input in UTF-8 mode ? */
 extern FLAG combining_screen;	/* UTF-8 combining character terminal ? */
 extern FLAG bidi_screen;	/* UTF-8 bidi terminal ? */
 extern FLAG joining_screen;	/* UTF-8 terminal with Arabic LAM/ALEF joining ? */
+extern FLAG halfjoining_screen;	/* Putty/mintty joining + spare space ? */
 
 extern FLAG mapped_term;	/* terminal in mapped 8-bit mode ? */
 
@@ -35,6 +36,7 @@ extern FLAG cjklow_term;	/* does CJK terminal support 81-9F range ? */
 extern int cjk_tab_width;	/* width of CJK TAB indicator */
 extern int cjk_lineend_width;	/* width of CJK line end indicator */
 
+
 /* basic data versions of Unicode (or in earlier cases, terminal-specific) */
 /* in addition to width_data_version, cjk_width_data_version is set 
    only in CJK legacy width mode (xterm -cjk_width);
@@ -43,11 +45,17 @@ extern int cjk_lineend_width;	/* width of CJK line end indicator */
    consistent with a specific Unicode version
  */
 typedef enum {
-	U0, U300beta, U300, U320beta, U320, U400, U410, U500, U510, 
+	U0, 
+	U300beta, U300, U320beta, U320, 
+	U400, U410, 
+	U500, U510, U520, 
+	U600, 
 	WIDTH_DATA_MAX
 } width_data_type;
 
-extern char * term_encoding_version_name _((int v));
+
+/* return name of Unicode version applied by terminal */
+extern char * term_Unicode_version_name _((int v));
 
 /* basic data versions of Unicode */
 extern int width_data_version;
@@ -69,6 +77,7 @@ extern int gnome_terminal_version;
 extern int rxvt_version;
 extern int konsole_version;
 extern int mintty_version;
+extern int mlterm_version;
 extern int screen_version;
 
 /* terminal features with respect to non-BMP characters (>= 0x10000) */
