@@ -20,22 +20,75 @@ FLAG tab_right = False;	/* Set if moving up/down on TAB should go left */
 
 #define dont_debug_move
 
-FLAG highlighting = True;
 
+/*
+ * Just count to determine the current line number.
+ */
 static
 void
-update_highlighting_marks (old_line, old_text, new_line, new_text)
-  LINE * old_line;
-  char * old_text;
-  LINE * new_line;
-  char * new_text;
+calc_line_no ()
 {
-  if (highlighting) {
-	
+  register LINE * line;
+  register int line_num = 0;
+
+  for (line = header->next; line != cur_line; line = line->next) {
+	line_num ++;
   }
+  line_number = line_num + 1;
+}
+
+/*
+ * Proceed returns the count'th line after 'line'. When count is negative
+ * it returns the count'th line before 'line'. When the next (previous)
+ * line is the tail (header) indicating EOF (TOF) it stops.
+ */
+LINE *
+proceed (line, count)
+  register LINE * line;
+  register int count;
+{
+  if (count < 0) {
+	while (count ++ < 0 && line != header) {
+		line = line->prev;
+	}
+  } else {
+	while (count -- > 0 && line != tail) {
+		line = line->next;
+	}
+  }
+  return line;
+}
+
+/*
+ * Reset assigns bot_line, top_line and cur_line according to 'head_line'
+ * which must be the first line of the screen, and a y-coordinate,
+ * which will be the current y-coordinate (if it isn't larger than last_y)
+ */
+void
+reset (head_line, screen_y)
+  LINE * head_line;
+  int screen_y;
+{
+  register LINE * line;
+
+  top_line = line = head_line;
+
+/* Search for bot_line (might be last line in file) */
+  for (last_y = 0; 
+	last_y < total_lines - 1 && last_y < SCREENMAX && line->next != tail; 
+	last_y ++
+      ) {
+	line = line->next;
+  }
+
+  bot_line = line;
+  y = (screen_y > last_y) ? last_y : screen_y;
+
+/* Set cur_line according to the new y value */
+  cur_line = proceed (top_line, y);
+  calc_line_no ();
 }
 
-int old_x = 0;	/* previous x position */
 
 static
 int
@@ -94,6 +147,8 @@ find_x (line, address)
   return x_in_line + x_left;
 }
 
+static int old_x = 0;	/* previous x position */
+
 /*
  * Find_address () returns the pointer in the line with given offset.
  * (Tabs are expanded).
@@ -112,7 +167,6 @@ find_address (line, new_x, cur_x)
   int tx = line->shift_count * - SHIFT_SIZE;
   char * prev_poi;
   int prev_x;
-  unsigned long unichar;
 
 #define navigate_tab_to_closer_side
 
@@ -156,10 +210,18 @@ find_address (line, new_x, cur_x)
 
 		advance_char_scr (& textp, & tx, line->text);
 
-		/* skip over combining characters */
-		if (combining_mode && encoding_has_combining ()) {
-			unichar = unicodevalue (textp);
-			while (iscombined_display (unichar, textp, line->text)) {
+		if (combining_mode) {
+			/* skip over combining characters */
+			unsigned long unichar = unicodevalue (textp);
+			while ((encoding_has_combining ()
+				&& iscombined_display (unichar, textp, line->text)
+			       )
+			       || isspacingcombining_unichar (unichar)
+			       || /* handle ARABIC TAIL FRAGMENT */
+			       (unichar == 0xFE73
+				&& iscombined_unichar (unichar, textp, line->text)
+			       )
+			      ) {
 				advance_char_scr (& textp, & tx, line->text);
 				unichar = unicodevalue (textp);
 			}
@@ -326,15 +388,18 @@ move_it (new_x, new_address, new_y)
 	}
   }
 
-/* Update internal highlighting marks */
-  update_highlighting_marks (cur_line, cur_text, line, new_address);
-
 /* Assign and position cursor */
   x = new_x;
   cur_text = new_address;
   cur_line = line;
-  calc_line_no ();
+
+/* Update selection highlighting */
+  update_selection_marks (x);
+  /* don't display_flush () here */
+
   line_number = lineno;
+  /* don't calc_line_no (), to prevent swapping */
+
   set_cursor_xy ();
 }
 
@@ -376,74 +441,6 @@ move_address_w_o_RD (nadd, ny)
 }
 
 
-/*
- * Just count to determine the current line number.
- */
-void
-calc_line_no ()
-{
-  register LINE * line;
-  register int line_num = 0;
-
-	for (line = header->next; line != cur_line; line = line->next) {
-		line_num ++;
-	}
-	line_number = line_num + 1;
-}
-
-/*
- * Proceed returns the count'th line after 'line'. When count is negative
- * it returns the count'th line before 'line'. When the next (previous)
- * line is the tail (header) indicating EOF (TOF) it stops.
- */
-LINE *
-proceed (line, count)
-  register LINE * line;
-  register int count;
-{
-  if (count < 0) {
-	while (count ++ < 0 && line != header) {
-		line = line->prev;
-	}
-  } else {
-	while (count -- > 0 && line != tail) {
-		line = line->next;
-	}
-  }
-  return line;
-}
-
-/*
- * Reset assigns bot_line, top_line and cur_line according to 'head_line'
- * which must be the first line of the screen, and a y-coordinate,
- * which will be the current y-coordinate (if it isn't larger than last_y)
- */
-void
-reset (head_line, screen_y)
-  LINE * head_line;
-  int screen_y;
-{
-  register LINE * line;
-
-  top_line = line = head_line;
-
-/* Search for bot_line (might be last line in file) */
-  for (last_y = 0; 
-	last_y < total_lines - 1 && last_y < SCREENMAX && line->next != tail; 
-	last_y ++
-      ) {
-	line = line->next;
-  }
-
-  bot_line = line;
-  y = (screen_y > last_y) ? last_y : screen_y;
-
-/* Set cur_line according to the new y value */
-  cur_line = proceed (top_line, y);
-  calc_line_no ();
-}
-
-
 /*
    Determine current character position within line (# characters).
  */
@@ -468,6 +465,22 @@ get_cur_col ()
   return cur_line->shift_count * SHIFT_SIZE + x;
 }
 
+/*
+   Return text column position within line.
+ */
+int
+get_text_col (line, textp)
+  LINE * line;
+  char * textp;
+{
+  char * tp = line->text;
+  int col = 0;
+  while (tp != textp && * tp != '\n' && * tp != '\0') {
+	advance_char_scr (& tp, & col, line->text);
+  }
+  return col;
+}
+
 
 /*======================================================================*\
 |*			Text/file information				*|
@@ -496,14 +509,21 @@ file_status (message, bytecount, charcount, filename, lines, textstat, writefl,
 {
   LINE * line;
   int line_num = 0;
-  char fileinfo [maxLINE_LEN];
-  char textinfo [maxLINE_LEN];
+  char fileinfo [maxPROMPTlen + maxFILENAMElen];
+  char flaginfo [maxPROMPTlen];
+  char textinfo [maxPROMPTlen];
   char * filesep = "";
 
   if (* message) {
 	filesep = " ";
   }
 
+  if (reading_pipe && * message) {
+	filename = "[standard input]";
+  } else if (! * filename) {
+	filename = "[no file]";
+  }
+
   /* determine information */
   if (bytecount < 0) {	/* # chars not given; count them */
 	bytecount = 0L;
@@ -534,16 +554,15 @@ file_status (message, bytecount, charcount, filename, lines, textstat, writefl,
 	}
   }
 
-  build_string (fileinfo, "%s%s%s%s%s%s, ",
-		message, filesep,
-		(reading_pipe && * message != '[') ?
-			"standard input" : filename,
-		viewing ? " (View only)" : "",
-		changed ? " (modified)" : "",
-		writefl ? "" : file_is_dir ? " (Directory)" : " (Readonly)");
+  build_string (fileinfo, "%s%s%s", message, filesep, filename);
+  build_string (flaginfo, "%s -%s%s ",
+		writefl ? "" : file_is_dir ? " (directory)" : " (readonly)",
+		viewing ? " (view only)" : "",
+		changed ? " (modified)" : "");
 
   if (textstat) {
-	build_string (textinfo, "line:char %d:%ld (total %d:%ld)%s%s",
+	build_string (textinfo, "%sline:char %d:%d (total %d:%ld)%s%s",
+		flaginfo,
 		line_number, get_cur_pos () + 1, lines, charcount,
 
 		(lines_per_page != 0) ? ", page " : "",
@@ -552,7 +571,8 @@ file_status (message, bytecount, charcount, filename, lines, textstat, writefl,
 			: ""
 		);
   } else {
-	build_string (textinfo, "lines:chars(bytes) %d:%ld(%ld)",
+	build_string (textinfo, "%slines:chars(bytes) %d:%ld(%ld)",
+		flaginfo,
 		lines,
 		charcount,
 		bytecount
